################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import random
from collections import defaultdict
from typing import Callable, Dict, List

import numpy as np

from .api import (
    CostFunction,
    NestedOptimizer,
    Optimizer,
    construct_history_info,
    optimization_result,
)
from .api.optimizer import extend_histories
from .history.recorder import RecorderFactory
from .history.recorder import recorder as _recorder


class MockOptimizer(Optimizer):
    def _minimize(
        self, cost_function, initial_params: np.ndarray, keep_history: bool = False
    ):
        new_parameters = initial_params
        for i in range(len(initial_params)):
            new_parameters[i] += random.random()
        new_parameters = np.array(new_parameters)
        return optimization_result(
            opt_value=cost_function(new_parameters),
            opt_params=new_parameters,
            nfev=1,
            **construct_history_info(cost_function, keep_history),
        )


def mock_cost_function(parameters: np.ndarray):
    return np.sum(parameters**2)


class MockNestedOptimizer(NestedOptimizer):
    """
    As most mock objects this implementation does not make much sense in itself,
    however it's an example of how a NestedOptimizer could be implemented.

    """

    @property
    def inner_optimizer(self) -> Optimizer:
        return self._inner_optimizer

    @property
    def recorder(self) -> RecorderFactory:
        return self._recorder

    def __init__(
        self,
        inner_optimizer: Optimizer,
        n_iters: int,
        recorder: RecorderFactory = _recorder,
    ):
        self._inner_optimizer = inner_optimizer
        self.n_iters = n_iters
        self._recorder = recorder

    def _minimize(
        self,
        cost_function_factory: Callable[[int], CostFunction],
        initial_params: np.ndarray,
        keep_history: bool = False,
    ):
        histories: Dict[str, List] = defaultdict(list)
        histories["history"] = []
        nfev = 0
        current_params = initial_params
        for i in range(self.n_iters):
            if i != 0:
                # Increase the length of params every iteration
                # and repeats optimization with the longer params vector.
                current_params = np.append(current_params, 1)

            # Cost function changes with every iteration of NestedOptimizer
            # because it's dependent on iteration number
            cost_function = cost_function_factory(i)
            if keep_history:
                cost_function = self.recorder(cost_function)
            opt_result = self.inner_optimizer.minimize(cost_function, initial_params)
            nfev += opt_result.nfev
            current_params = opt_result.opt_params
            if keep_history:
                histories = extend_histories(cost_function, histories)  # type: ignore
        return optimization_result(
            opt_value=opt_result.opt_value,
            opt_params=current_params,
            nit=self.n_iters,
            nfev=nfev,
            **histories,
        )



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from typing import Tuple, Union

import numpy as np
from dimod import BinaryQuadraticModel


def evaluate_bitstring_for_qubo(
    bitstring: Union[str, np.ndarray, Tuple[int, ...]], qubo: BinaryQuadraticModel
):
    """Returns the energy associated with given bitstring for a specific qubo.

    Args:
        bitstring: string/array of zeros and ones representing solution to a qubo .
        qubo: qubo that we want to evaluate for.

    Returns:
        float: energy associated with a bistring
    """
    return qubo.energy({i: bit for i, bit in enumerate(map(int, bitstring))})



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import logging

import numpy as np
import pytest

from orquestra.opt.api import always, every_nth
from orquestra.opt.cost_function_augmentations import (
    ConditionalSideEffect,
    augment_cost_function,
    function_logger,
)


class Example:
    def __init__(self, function, label):
        self.function = function
        self.label = label

    def __call__(self, params):
        return self.function(params)


def augmentation_1(function):
    return Example(function, "example_1")


def augmentation_2(function):
    return Example(function, "example_2")


class AugmentationWithNumber:
    def __init__(self, function):
        self.number = 100
        self.function = function

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)


def foo(x):
    return x**2


class Foo:
    def __init__(self):
        self.number = 10
        self.string = "test"

    def __call__(self, x):
        return x**2


class FooGradient:
    attr = "custom-attr"

    def __call__(self, *args, **kwargs):
        return np.array([1.0, 2.0, 3.0])


class FooWithGradient:
    def __init__(self):
        self.gradient = FooGradient()

    def __call__(self, x):
        return x**3


class ExampleSideEffect(ConditionalSideEffect):
    def _act(self, result, params):
        print(f"{self.call_number} {result} {params}")


def example_side_effect_augmentation(predicate):
    def _augment(function):
        return ExampleSideEffect(function, predicate)

    return _augment


class TestAugmentingCostFunction:
    def test_returns_the_same_function_if_no_augmentations_are_provided(self):
        assert augment_cost_function(foo) == foo

    def test_applies_augmentation_to_cost_function_in_correct_order(self):
        augmented_foo = augment_cost_function(foo, [augmentation_1, augmentation_2])

        assert augmented_foo.label == "example_2"
        assert augmented_foo.function.label == "example_1"
        assert augmented_foo.function.function == foo

    class TestQueryingAttributes:
        def test_cost_function_attributes_are_propagated(self):
            augmented_foo = augment_cost_function(
                Foo(), [augmentation_1, augmentation_2]
            )
            assert augmented_foo.number == 10
            assert augmented_foo.string == "test"

        def test_augmentation_attributes_override_cost_function_attributes(self):
            augmented_foo = augment_cost_function(Foo(), [AugmentationWithNumber])
            assert augmented_foo.number == 100
            assert augmented_foo.string == "test"

        def test_gradient_attributes_are_propagated(self):
            augmented_foo = augment_cost_function(
                FooWithGradient(),
                gradient_augmentations=[augmentation_1, augmentation_2],
            )

            assert augmented_foo.gradient.attr == "custom-attr"

    class TestSettingAttributes:
        def test_attribute_is_set_on_function_if_it_is_not_present_on_augmentation(
            self,
        ):
            cost_function = Foo()
            augmented_foo = augment_cost_function(cost_function, [augmentation_1])

            augmented_foo.number = 100

            assert augmented_foo.number == 100
            assert cost_function.number == 100

        def test_attribute_is_set_only_on_augmentation_if_it_is_already_present(self):
            cost_function = Foo()
            augmented_foo = augment_cost_function(
                cost_function, [AugmentationWithNumber]
            )

            augmented_foo.number = 1000

            assert augmented_foo.number == 1000
            assert cost_function.number == 10

        def test_attribute_is_set_on_gradient_if_it_is_not_present_on_its_augmentation(
            self,
        ):
            cost_function = FooWithGradient()
            augmented_foo = augment_cost_function(cost_function, [augmentation_1])

            augmented_foo.gradient.attr = "test"

            assert augmented_foo.gradient.attr == "test"
            assert cost_function.gradient.attr == "test"


class TestConditionalSideEffect:
    def test_passes_through_value_returned_by_wrapped_function(self):
        augmented_foo = augment_cost_function(
            foo, [example_side_effect_augmentation(always)]
        )

        assert augmented_foo(10) == foo(10)

    def test_fires_only_when_the_predicate_is_true(self, capsys):
        augmented_foo = augment_cost_function(
            foo, [example_side_effect_augmentation(every_nth(2))]
        )

        augmented_foo(3)
        augmented_foo(4)
        augmented_foo(10)
        augmented_foo(5)

        assert capsys.readouterr().out == "0 9 3\n2 100 10\n"


class TestLoggingAugmentation:
    @pytest.mark.parametrize("level", [logging.INFO, logging.WARNING, logging.ERROR])
    def test_logs_with_correct_level(self, level, caplog):
        augmented_foo = augment_cost_function(foo, [function_logger(level=level)])

        with caplog.at_level(level):
            augmented_foo(3)

        assert "Function called. Call number: 0, function value: 9" in caplog.text

    def test_uses_provided_logger(self, caplog):
        logger = logging.getLogger("my.logger")
        augmented_foo = augment_cost_function(
            foo, [function_logger(level=logging.WARNING, logger=logger)]
        )

        with caplog.at_level(logging.WARNING, "my.logger"):
            augmented_foo(4)

        assert "Function called. Call number: 0, function value: 16" in caplog.text



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import numpy as np
import pytest

from orquestra.opt.api import FunctionWithGradient
from orquestra.opt.api.optimizer_test import OptimizerTests
from orquestra.opt.gradients import finite_differences_gradient
from orquestra.opt.optimizers.scipy_optimizer import ScipyOptimizer


@pytest.fixture(
    params=[
        {"method": "BFGS"},
        {"method": "L-BFGS-B"},
        {"method": "Nelder-Mead"},
        {"method": "SLSQP"},
        {"method": "COBYLA", "options": {"maxiter": 50000, "tol": 1e-7}},
    ]
)
def optimizer(request):
    return ScipyOptimizer(**request.param)


@pytest.fixture(
    params=[
        {"method": "L-BFGS-B"},
        {"method": "Nelder-Mead"},
        {"method": "SLSQP"},
    ]
)
def optimizer_with_bounds(request):
    bounds = [(2, 3), (2, 3)]
    return ScipyOptimizer(bounds=bounds, **request.param)


@pytest.fixture(params=[True, False])
def keep_history(request):
    return request.param


class TestScipyOptimizer(OptimizerTests):
    def test_optimizers_work_with_bounds_provided(
        self, optimizer_with_bounds, sum_x_squared
    ):

        # Given
        cost_function = FunctionWithGradient(
            sum_x_squared, finite_differences_gradient(sum_x_squared)
        )

        initial_params = np.array([2.5, 2.5])
        target_params = np.array([2, 2])
        target_value = 8

        # When
        results = optimizer_with_bounds.minimize(
            cost_function, initial_params=initial_params
        )

        # Then
        assert results.opt_value == pytest.approx(target_value, abs=1e-3)
        assert results.opt_params == pytest.approx(target_params, abs=1e-3)

    def test_SLSQP_with_equality_constraints(self, sum_x_squared, rosenbrock_function):
        # Given
        cost_function = FunctionWithGradient(
            rosenbrock_function, finite_differences_gradient(rosenbrock_function)
        )
        constraint_cost_function = sum_x_squared

        constraints = ({"type": "eq", "fun": constraint_cost_function},)
        optimizer = ScipyOptimizer(method="SLSQP", constraints=constraints)
        initial_params = np.array([1, 1])
        target_params = np.array([0, 0])
        target_value = 1

        # When
        results = optimizer.minimize(cost_function, initial_params=initial_params)

        # Then
        assert results.opt_value == pytest.approx(target_value, abs=1e-3)
        assert results.opt_params == pytest.approx(target_params, abs=1e-3)

    def test_SLSQP_with_inequality_constraints(self, rosenbrock_function):
        # Given
        cost_function = FunctionWithGradient(
            rosenbrock_function, finite_differences_gradient(rosenbrock_function)
        )
        constraints = {"type": "ineq", "fun": lambda x: x[0] + x[1] - 3}
        optimizer = ScipyOptimizer(method="SLSQP")
        initial_params = np.array([0, 0])

        # When
        results_without_constraints = optimizer.minimize(
            cost_function, initial_params=initial_params
        )
        optimizer.constraints = constraints
        results_with_constraints = optimizer.minimize(
            cost_function, initial_params=initial_params
        )

        # Then
        assert results_without_constraints.opt_value == pytest.approx(
            results_with_constraints.opt_value, abs=1e-1
        )
        assert results_with_constraints.opt_params.sum() >= 3



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""Test cases for recording basic functions."""
from unittest.mock import Mock, call

import numpy as np
import pytest

from orquestra.opt.api.example_functions import (
    Function2,
    Function5,
    function_1,
    sum_of_squares,
)
from orquestra.opt.api.save_conditions import SaveCondition, every_nth
from orquestra.opt.history import recorder


@pytest.mark.parametrize(
    "source_function,param",
    [
        (np.sin, 10),
        (sum_of_squares, [1, 2, 3]),
        (function_1, np.array([2, 3])),
        (Function2(5), np.array([1, 2, 3])),
        (Function5(2), np.array([1, 2, 3])),
    ],
)
def test_recorder_propagates_calls_to_wrapped_function(source_function, param):
    function = recorder(source_function)
    assert function(param) == source_function(param)


@pytest.mark.parametrize(
    "source_function,params_sequence",
    [
        (np.exp, [1, 3, 5, 10]),
        (sum_of_squares, [[1, 2, 3], [4, 5, 6]]),
        (function_1, [np.array([-2, -3]), np.array([0, 1])]),
        (Function2(5), [np.array([1, 0, -1]), np.array([0, 1, 2])]),
        (Function5(2), [np.array([1, 2, 3]), np.array([0, 2, 4])]),
    ],
)
def test_recorder_does_not_make_any_redundant_calls_to_wrapped_function(
    source_function, params_sequence
):
    spy = Mock(wraps=source_function)
    function = recorder(spy)

    for params in params_sequence:
        function(params)

    assert spy.call_args_list == [call(params) for params in params_sequence]


@pytest.mark.parametrize(
    "source_function,params_sequence",
    [
        (np.exp, [1, 3, 5, 10]),
        (sum_of_squares, [[1, 2, 3], [4, 5, 6]]),
        (function_1, [np.array([-2, -3]), np.array([0, 1]), np.array([1, 2])]),
        (
            Function2(5),
            [
                np.array(args)
                for args in [[-1, 0, 1], [10, 20, 30], [0, 1, 2], [3, 4, 5]]
            ],
        ),
        (Function5(2), [np.array([1, 2, 3]), np.array([0, 2, 4])]),
    ],
)
def test_by_default_recorder_records_all_evaluations(source_function, params_sequence):
    function = recorder(source_function)

    for params in params_sequence:
        function(params)

    np.testing.assert_array_equal(
        [entry.params for entry in function.history], params_sequence
    )
    assert [entry.value for entry in function.history] == [
        source_function(params) for params in params_sequence
    ]
    assert [entry.call_number for entry in function.history] == list(
        range(len(params_sequence))
    )


@pytest.mark.parametrize(
    "source_function,params_sequence,condition",
    [
        (sum_of_squares, [list(range(n)) for n in range(100)], every_nth(5)),
        (function_1, [np.array([k, k + 1]) for k in range(1000)], every_nth(101)),
        (
            Function2(10),
            [np.array([k, 2 * k, 3 * k]) for k in range(200)],
            every_nth(21),
        ),
        (
            Function5(2),
            [np.array([k, 2 * k, 3 * k]) for k in range(200)],
            every_nth(21),
        ),
    ],
)
def test_recorder_records_only_calls_for_which_save_condition_evaluates_to_true(
    source_function, params_sequence, condition: SaveCondition
):
    function = recorder(source_function, save_condition=condition)
    expected_values, expected_params, expected__call_numbers = zip(
        *filter(
            lambda x: condition(*x),
            (
                (source_function(params), params, i)
                for i, params in enumerate(params_sequence)
            ),
        )
    )
    for params in params_sequence:
        function(params)

    assert [entry.call_number for entry in function.history] == list(
        expected__call_numbers
    )
    np.testing.assert_array_equal(
        [entry.params for entry in function.history], list(expected_params)
    )
    assert [entry.value for entry in function.history] == list(expected_values)


@pytest.mark.parametrize(
    "source_function,params_sequence",
    [
        (sum_of_squares, [[1, 2, 3], [4, 5, 6]]),
        (function_1, [np.array([-2, -3]), np.array([0, 1]), np.array([1, 2])]),
        (
            Function2(5),
            [
                np.array(args)
                for args in [[-1, 0, 1], [10, 20, 30], [0, 1, 2], [3, 4, 5]]
            ],
        ),
        (Function5(2), [np.array([1, 2, 3]), np.array([0, 2, 4])]),
    ],
)
def test_recorder_stores_copy_of_parameters(source_function, params_sequence):
    function = recorder(source_function)

    for params in params_sequence:
        function(params)

    for original_params, entry in zip(params_sequence, function.history):
        np.testing.assert_array_equal(original_params, entry.params)
        assert original_params is not entry.params



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from typing import Callable, Dict, Optional, Union

import cma
import numpy as np
from scipy.optimize import OptimizeResult

from ..api import (
    CallableWithGradient,
    Optimizer,
    construct_history_info,
    optimization_result,
)
from ..history.recorder import RecorderFactory
from ..history.recorder import recorder as _recorder


class CMAESOptimizer(Optimizer):
    def __init__(
        self,
        sigma_0: float,
        options: Optional[Dict] = None,
        recorder: RecorderFactory = _recorder,
    ):
        """
        Integration with CMA-ES optimizer: https://github.com/CMA-ES/pycma .
        Args:
            sigma_0: please refer to https://github.com/CMA-ES/pycma documentation.
            options: dictionary with options for the optimizer,
                please refer to https://github.com/CMA-ES/pycma documentation.
            recorder: recorder object which defines how to store
                the optimization history.
        """
        super().__init__(recorder=recorder)
        self.sigma_0 = sigma_0
        if options is None:
            options = {}
        self.options = options

    def _minimize(
        self,
        cost_function: Union[CallableWithGradient, Callable],
        initial_params: np.ndarray,
        keep_history: bool = False,
    ) -> OptimizeResult:
        """Minimize using the Covariance Matrix Adaptation Evolution Strategy
        (CMA-ES).

        Note:
            Original CMA-ES implementation stores optimization history by default.
            This is a separate mechanism from the one controlled by recorder, and
            therefore is turned on even if keep_history is set to false, which might
            lead to memory issues in some extreme cases.
            However, we expose only the recording performed using provided recorder.

        Args:
            cost_function: object representing cost function we want to minimize
            initial_params: initial guess for the ansatz parameters.
            keep_history: flag indicating whether history of cost function
                evaluations should be recorded.
        """
        strategy = cma.CMAEvolutionStrategy(initial_params, self.sigma_0, self.options)
        result = strategy.optimize(cost_function).result

        return optimization_result(
            opt_value=result.fbest,
            opt_params=result.xbest,
            nfev=result.evaluations,
            nit=result.iterations,
            cma_xfavorite=list(result.xfavorite),
            **construct_history_info(cost_function, keep_history)  # type: ignore
        )



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import numpy as np
import pytest

from orquestra.opt.api import function_with_gradient
from orquestra.opt.api.optimizer_test import OptimizerTests
from orquestra.opt.gradients import finite_differences_gradient
from orquestra.opt.optimizers.simple_gradient_descent import SimpleGradientDescent


@pytest.fixture(
    params=[
        {"learning_rate": 0.1, "number_of_iterations": 100},
        {"learning_rate": 0.15, "number_of_iterations": 100},
        {"learning_rate": 0.215242, "number_of_iterations": 100},
        {"learning_rate": 0.99, "number_of_iterations": 1000},
    ]
)
def optimizer(request):
    return SimpleGradientDescent(**request.param)


@pytest.fixture(params=[True, False])
def keep_history(request):
    return request.param


class TestSimpleGradientDescent(OptimizerTests):
    @pytest.fixture
    def sum_x_squared(self):
        def _sum_x_squared(x):
            return sum(x**2)

        return function_with_gradient(
            _sum_x_squared, finite_differences_gradient(_sum_x_squared)
        )

    def test_optimizer_succeeds_with_optimizing_rosenbrock_function(
        self, optimizer, rosenbrock_function, keep_history
    ):
        pytest.xfail(
            """This test fails since the gradient of the rosenbrock function "
            "is too sensitive when using finite differences"""
        )

    def test_optimizer_succeeds_on_cost_function_without_gradient(
        self, optimizer, sum_x_squared
    ):
        pytest.xfail(
            """This test fails since TestSimpleGradientDescent requires cost_function "
            "to have gradient method"""
        )

    def test_fails_to_initialize_when_number_of_iterations_is_negative(self):
        with pytest.raises(AssertionError):
            SimpleGradientDescent(0.1, -1)

    def test_fails_to_minimize_when_cost_function_does_not_have_gradient_method(
        self, optimizer
    ):
        def cost_function(x):
            return sum(x)

        with pytest.raises(AssertionError):
            optimizer.minimize(cost_function, np.array([0, 0]))

    def test_history_contains_function_evaluations(self, optimizer, sum_x_squared):
        results = optimizer.minimize(sum_x_squared, np.array([1, 0]), keep_history=True)

        assert len(results.history) == optimizer.number_of_iterations



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from ..api._problem_evaluation import solve_problem_by_exhaustive_search
from .generators import get_random_ising_hamiltonian
from .graph_partition import GraphPartitioning
from .max_independent_set import MaxIndependentSet
from .maxcut import MaxCut
from .vertex_cover import VertexCover



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import pytest

from orquestra.opt.api.example_functions import (
    Function2,
    Function5,
    function_1,
    function_3,
    function_4,
    function_6,
)
from orquestra.opt.history import recorder


@pytest.mark.parametrize("func", [Function2(5), function_3, function_4, Function5(0.5)])
def test_recorder_correctly_gets_attributes_of_recorder_function(func):
    func.custom_attribute = "custom-attr"

    recorded = recorder(func)

    assert recorded.custom_attribute == func.custom_attribute

    func.custom_attribute = "some-other-value"

    assert recorded.custom_attribute == func.custom_attribute


@pytest.mark.parametrize("func", [Function2(5), function_3, function_4, Function5(0.5)])
def test_recorder_correctly_sets_attributes_of_recorder_function(func):
    func.custom_attribute = "custom-attr"

    recorded = recorder(func)

    recorded.custom_attribute = "new-value"

    assert func.custom_attribute == "new-value"


@pytest.mark.parametrize("func", [function_1, function_6])
def test_function_with_gradient_gets_wrapped_function_attributes(func):
    func.function.test = "test-value"

    assert func.test == "test-value"


@pytest.mark.parametrize("func", [function_1, function_6])
def test_function_with_gradient_sets_function_attributes(func):
    func.test = "test-value"

    assert func.function.test == "test-value"



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import networkx as nx
from orquestra.quantum.openfermion import QubitOperator

from ..api.problem import Problem


class MaxCut(Problem):
    """Solves max cut problem on an undirected weighted graph using an ising model
    formulation.

    The solution of a maxcut problem is the set of nodes S that maximizes the
    wieght of the edges connecting the nodes in S to the rest of the graph.
    From "A Quantum Approximate Optimization Algorithm" by E. Farhi, eq. 12
    (https://arxiv.org/pdf/1411.4028.pdf)
    and
    "Performance of the Quantum Approximate Optimization Algorithm on the Maximum
    Cut Problem" eq. 1 (https://arxiv.org/pdf/1811.08419.pdf).
    """

    def _build_hamiltonian(self, graph: nx.Graph) -> QubitOperator:
        """Converts a MAXCUT instance, as described by a weighted graph, to an Ising
        Hamiltonian. It allows for different convention in the choice of the
        Hamiltonian. The returned Hamiltonian is consistent with the definitions from
        "A Quantum Approximate Optimization Algorithm" by E. Farhi, eq. 12
        (https://arxiv.org/pdf/1411.4028.pdf)
        and
        "Performance of the Quantum Approximate Optimization Algorithm on the Maximum
        Cut Problem" eq. 1 (https://arxiv.org/pdf/1811.08419.pdf).

        Note: In the convention we assumed, values of the cuts are negative
        to frame the problem as a minimization problem.
        So for a linear graph 0--1--2 with weights all equal 1, and the solution
        [0,1,0], the returned value will be equal to -2.

        Args:
            graph: undirected weighted graph defining the problem
        """
        hamiltonian = QubitOperator()
        shift = 0.0

        for i, j in graph.edges:
            try:
                weight = graph.adj[i][j]["weight"]
            except KeyError:
                weight = 1

            hamiltonian += weight * QubitOperator(f"Z{i} Z{j}")
            shift -= weight

        return 0.5 * (hamiltonian + shift)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
# TODO : Remove everythin in this file
import itertools
import random
import warnings
from typing import Iterator, Optional, Sequence, TypeVar

import networkx as nx


def compare_graphs(graph1: nx.Graph, graph2: nx.Graph) -> bool:
    """Compares two NetworkX graph objects to see if they are identical.
    NOTE: this is *not* solving isomorphism problem.
    """

    for n1, n2 in zip(graph1.nodes, graph2.nodes):
        if n1 != n2:
            return False
    for e1, e2 in zip(graph1.edges, graph2.edges):
        if e1 != e2:
            return False
    return True


def generate_graph_node_dict(graph: nx.Graph) -> dict:
    """Generates a dictionary containing key:value pairs in the form of
    nx.Graph node : integer index of the node
    """
    nodes_int_map = []
    for node_index, node in enumerate(graph.nodes):
        nodes_int_map.append((node, node_index))
    nodes_dict = dict(nodes_int_map)
    return nodes_dict


T = TypeVar("T")


Sampler = Iterator[T]


def uniform_sampler(min_value: float = 0.0, max_value: float = 1.0) -> Sampler[float]:
    while True:
        yield random.uniform(min_value, max_value)


def constant_sampler(value: T) -> Sampler[T]:
    return itertools.repeat(value)


def choice_sampler(choices: Sequence[T]) -> Sampler[T]:
    while True:
        yield random.choice(choices)


def normal_sampler(mu: float = 0.0, sigma: float = 1.0) -> Sampler[float]:
    while True:
        yield random.normalvariate(mu, sigma)


def generate_random_graph_erdos_renyi(
    num_nodes: int,
    edge_probability: float,
    weight_sampler: Sampler = constant_sampler(1.0),
    seed: Optional[int] = None,
) -> nx.Graph:
    """Randomly generate a graph from Erdos-Renyi ensemble.  A graph is constructed by
    connecting nodes randomly.  Each edge is included in the graph with probability p
    independent from every other edge. Equivalently, all graphs with n nodes and M edges
    have equal probability.
    Args:
        num_nodes: Number of nodes in the result graph.
        edge_probability: Probability of connecting two nodes.
        weight_sampler: Used to sample edge weights. Defaults `static_sampler`,
            i.e. all edge weights are set to 1.0.
        seed: if provided, sets the global seed
    """
    output_graph = nx.erdos_renyi_graph(n=num_nodes, p=edge_probability, seed=seed)
    _weight_graph_edges(output_graph, weight_sampler, seed)

    return output_graph


def generate_random_regular_graph(
    num_nodes: int,
    degree: int,
    weight_sampler: Sampler = constant_sampler(1.0),
    seed: Optional[int] = None,
) -> nx.Graph:
    """Randomly generate a d-regular graph.
    A graph is generated by picking uniformly a graph among the set of graphs
    with the desired number of nodes and degree.
    Args:
        num_nodes: Number of nodes in the generated graph.
        degree: Degree of each edge.
        weight_sampler: Used to sample edge weights. Defaults `static_sampler`,
            i.e. all edge weights are set to 1.0.
        seed: if provided, sets the global seed
    """
    output_graph = nx.random_regular_graph(d=degree, n=num_nodes, seed=seed)
    _weight_graph_edges(output_graph, weight_sampler, seed)

    return output_graph


def generate_caveman_graph(
    number_of_cliques: int,
    size_of_cliques: int,
    weight_sampler: Sampler = constant_sampler(1.0),
    seed: Optional[int] = None,
) -> nx.Graph:
    output_graph = nx.caveman_graph(number_of_cliques, size_of_cliques)
    _weight_graph_edges(output_graph, weight_sampler, seed)
    return output_graph


def generate_ladder_graph(
    length_of_ladder: int,
    weight_sampler: Sampler = constant_sampler(1.0),
    seed: Optional[int] = None,
) -> nx.Graph:
    graph = nx.ladder_graph(length_of_ladder)
    _weight_graph_edges(graph, weight_sampler, seed)
    return graph


def generate_barbell_graph(
    number_of_vertices_complete_graph: int,
    weight_sampler: Sampler = constant_sampler(1.0),
    seed: Optional[int] = None,
) -> nx.Graph:
    graph = nx.barbell_graph(number_of_vertices_complete_graph, 0)
    _weight_graph_edges(graph, weight_sampler, seed)

    return graph


def _weight_graph_edges(
    graph: nx.Graph,
    sampler: Sampler,
    seed: Optional[int] = None,
) -> nx.Graph:
    """Update the weights of all the edges of a graph in place.
    Args:
        graph: The graph to mutate.
        sampler:
    """
    if graph.is_multigraph():
        raise ValueError("Cannot deal with multigraphs")

    if seed is not None:
        random.seed(seed)

    weighted_edges = [(e[0], e[1], next(sampler)) for e in graph.edges]

    # If edges already present, it will effectively update them (except for multigraph)
    graph.add_weighted_edges_from(weighted_edges)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import copy

import networkx as nx
import pytest

from orquestra.opt.problems import VertexCover

from ._helpers import graph_node_index, make_graph

MONOTONIC_GRAPH_OPERATOR_TERM_PAIRS = [
    (
        make_graph(node_ids=range(2), edges=[(0, 1)]),
        {
            (): 2.25,
            ((0, "Z"),): -0.75,
            ((1, "Z"),): -0.75,
            ((0, "Z"), (1, "Z")): 1.25,
        },
    ),
    (
        make_graph(node_ids=range(3), edges=[(0, 1), (0, 2)]),
        {
            (): 4,
            ((0, "Z"),): -2,
            ((1, "Z"),): -0.75,
            ((2, "Z"),): -0.75,
            ((0, "Z"), (1, "Z")): 1.25,
            ((0, "Z"), (2, "Z")): 1.25,
        },
    ),
    (
        make_graph(node_ids=range(4), edges=[(0, 1), (0, 2), (0, 3)]),
        {
            (): 5.75,
            ((0, "Z"),): -3.25,
            ((1, "Z"),): -0.75,
            ((2, "Z"),): -0.75,
            ((3, "Z"),): -0.75,
            ((0, "Z"), (1, "Z")): 1.25,
            ((0, "Z"), (2, "Z")): 1.25,
            ((0, "Z"), (3, "Z")): 1.25,
        },
    ),
    (
        make_graph(node_ids=range(5), edges=[(0, 1), (1, 2), (3, 4)]),
        {
            (): 6.25,
            ((0, "Z"),): -0.75,
            ((1, "Z"),): -2,
            ((2, "Z"),): -0.75,
            ((3, "Z"),): -0.75,
            ((4, "Z"),): -0.75,
            ((0, "Z"), (1, "Z")): 1.25,
            ((1, "Z"), (2, "Z")): 1.25,
            ((3, "Z"), (4, "Z")): 1.25,
        },
    ),
]

NONMONOTONIC_GRAPH_OPERATOR_TERM_PAIRS = [
    (
        make_graph(node_ids=[4, 2], edges=[(2, 4)]),
        {
            (): 2.25,
            ((0, "Z"),): -0.75,
            ((1, "Z"),): -0.75,
            ((0, "Z"), (1, "Z")): 1.25,
        },
    ),
    (
        make_graph(node_ids="CBA", edges=[("C", "B"), ("C", "A")]),
        {
            (): 4,
            ((0, "Z"),): -2,
            ((1, "Z"),): -0.75,
            ((2, "Z"),): -0.75,
            ((0, "Z"), (1, "Z")): 1.25,
            ((0, "Z"), (2, "Z")): 1.25,
        },
    ),
]

GRAPH_EXAMPLES = [
    *[graph for graph, _ in MONOTONIC_GRAPH_OPERATOR_TERM_PAIRS],
    *[graph for graph, _ in NONMONOTONIC_GRAPH_OPERATOR_TERM_PAIRS],
    make_graph(
        node_ids=range(10),
        edges=[
            (0, 2),
            (0, 3),
            (1, 2),
            (4, 5),
            (0, 8),
        ],
    ),
    make_graph(
        node_ids=["foo", "bar", "baz"],
        edges=[
            ("foo", "baz"),
            ("bar", "baz"),
        ],
    ),
]


GRAPH_SOLUTION_COST_LIST = [
    (make_graph(node_ids=range(2), edges=[(0, 1)]), [0, 0], 2),
    (make_graph(node_ids=range(2), edges=[(0, 1)]), [0, 1], 1),
    (
        make_graph(
            node_ids=range(4), edges=[(0, 1, 1), (0, 2, 2), (0, 3, 3)], use_weights=True
        ),
        [1, 0, 0, 0],
        3,
    ),
    (make_graph(node_ids=range(4), edges=[(0, 1), (0, 2), (0, 3)]), [0, 0, 1, 1], 2),
    (make_graph(node_ids=range(4), edges=[(0, 1), (0, 2), (0, 3)]), [0, 1, 1, 1], 1),
    (
        make_graph(node_ids=range(5), edges=[(0, 1), (1, 2), (3, 4)]),
        [1, 1, 1, 1, 1],
        15,
    ),
]

GRAPH_BEST_SOLUTIONS_COST_LIST = [
    (make_graph(node_ids=range(2), edges=[(0, 1)]), [(0, 1), (1, 0)], 1),
    (
        make_graph(node_ids=range(3), edges=[(0, 1), (0, 2)]),
        [(0, 1, 1)],
        1,
    ),
    (
        make_graph(node_ids=range(4), edges=[(0, 1), (0, 2), (0, 3)]),
        [
            (0, 1, 1, 1),
        ],
        1,
    ),
    (
        make_graph(node_ids=range(5), edges=[(0, 1), (1, 2), (3, 4)]),
        [(1, 0, 1, 0, 1), (1, 0, 1, 1, 0)],
        2,
    ),
]


class TestGetVertexCoverHamiltonian:
    @pytest.mark.parametrize(
        "graph,terms",
        [
            *MONOTONIC_GRAPH_OPERATOR_TERM_PAIRS,
            *NONMONOTONIC_GRAPH_OPERATOR_TERM_PAIRS,
        ],
    )
    def test_returns_expected_terms(self, graph, terms):
        qubit_operator = VertexCover().get_hamiltonian(graph)
        assert qubit_operator.terms == terms

    @pytest.mark.parametrize("graph", GRAPH_EXAMPLES)
    def test_has_1_25_weight_on_edge_terms(self, graph: nx.Graph):
        qubit_operator = VertexCover().get_hamiltonian(graph)

        for vertex_id1, vertex_id2 in graph.edges:
            qubit_index1 = graph_node_index(graph, vertex_id1)
            qubit_index2 = graph_node_index(graph, vertex_id2)
            assert (
                qubit_operator.terms[((qubit_index1, "Z"), (qubit_index2, "Z"))] == 1.25
            )

    @pytest.mark.parametrize("graph", GRAPH_EXAMPLES)
    def test_has_correct_constant_term(self, graph: nx.Graph):
        expected_constant_term = 0.0

        qubit_operator = VertexCover().get_hamiltonian(graph)

        expected_constant_term += (5 / 4) * len(graph.edges)
        expected_constant_term += 0.5 * len(graph.nodes)

        assert qubit_operator.terms[()] == expected_constant_term


class TestEvaluateVertexCoverSolution:
    @pytest.mark.parametrize("graph,solution,target_value", [*GRAPH_SOLUTION_COST_LIST])
    def test_evaluate_vertex_cover_solution(self, graph, solution, target_value):
        value = VertexCover().evaluate_solution(solution, graph)
        assert value == target_value

    @pytest.mark.parametrize("graph,solution,target_value", [*GRAPH_SOLUTION_COST_LIST])
    def test_evaluate_vertex_cover_solution_with_invalid_input(
        self, graph, solution, target_value
    ):
        too_long_solution = solution + [1]
        too_short_solution = solution[:-1]
        invalid_value_solution = copy.copy(solution)
        invalid_value_solution[0] = -1
        invalid_solutions = [
            too_long_solution,
            too_short_solution,
            invalid_value_solution,
        ]
        for invalid_solution in invalid_solutions:
            with pytest.raises(ValueError):
                _ = VertexCover().evaluate_solution(invalid_solution, graph)


class TestSolveVertexCoverByExhaustiveSearch:
    @pytest.mark.parametrize(
        "graph,target_solutions,target_value", [*GRAPH_BEST_SOLUTIONS_COST_LIST]
    )
    def test_solve_vertex_cover_by_exhaustive_search(
        self, graph, target_solutions, target_value
    ):
        value, solutions = VertexCover().solve_by_exhaustive_search(graph)
        assert set(solutions) == set(target_solutions)
        assert value == target_value



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import json
from typing import Any, Dict

import dimod
from orquestra.quantum.typing import DumpTarget, LoadSource
from orquestra.quantum.utils import ensure_open


def bqm_to_serializable(bqm: dimod.BinaryQuadraticModel) -> Dict[str, Any]:
    """Convert binary quadratic model to a serializable dictionary.

    Args:
        bqm: binary quadratic model to convert

    Returns:
        Dictionary with the following keys:
        - linear: list of pairs (i, a_i), where a_i is linear coefficient
          corresponding to variable i,
        - quadratic: list of triples (i, j, b_ij) where b_ij is quadratic
          coefficient corresponding to variables i < j,
        - offset: offset taken from the model.
        - vartype: field determining whether variables in the model are
          from the set {-1, 1} ("SPIN") or the set {0, 1} ("BINARY").

    Notes:
        A list of tuples was chosen instead of dictionary because dictionary
        keyed with tuples is not JSON-serializable.
    """
    return {
        "linear": [(label, coef) for label, coef in bqm.linear.items()],
        "quadratic": [
            (*sorted((label_1, label_2)), coef)
            for (label_1, label_2), coef in bqm.quadratic.items()
        ],
        "offset": bqm.offset,
        "vartype": bqm.vartype.name,
    }


def bqm_from_serializable(serializable: Dict[str, Any]) -> dimod.BinaryQuadraticModel:
    """Create a binary quadratic model from serializable dictionary.

    Args:
        serializable: dictionary representing BQM. The expected format
        of this dictionary as the same as the output format of
        `bqm_to_serializable`.

    Returns:
        Binary quadratic model converted from the input dictionary.
    """

    def ensure_hashable(i):
        return tuple(i) if isinstance(i, list) else i

    return dimod.BinaryQuadraticModel(
        {ensure_hashable(i): coef for i, coef in serializable["linear"]},
        {
            (ensure_hashable(i), ensure_hashable(j)): coef
            for i, j, coef in serializable["quadratic"]
        },
        serializable["offset"],
        vartype=serializable["vartype"],
    )


def load_qubo(input_file: LoadSource):

    with ensure_open(input_file, "r") as f:
        qubo_dict = json.load(f)

    return bqm_from_serializable(qubo_dict)


def save_qubo(qubo, output_file: DumpTarget):
    qubo_dict = bqm_to_serializable(qubo)

    with ensure_open(output_file, "w") as f:
        json.dump(qubo_dict, f)


def save_sampleset(sampleset, output_file: DumpTarget):
    sampleset_dict = sampleset.to_serializable()

    with ensure_open(output_file, "w") as f:
        json.dump(sampleset_dict, f)


def load_sampleset(input_file: LoadSource):

    with ensure_open(input_file, "r") as f:
        sampleset_dict = json.load(f)

    return dimod.SampleSet.from_serializable(sampleset_dict)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from abc import ABC, abstractmethod
from typing import Callable, Dict, List, Union, cast

import numpy as np
import scipy
from scipy.optimize import OptimizeResult

from ..history.recorder import (
    AnyHistory,
    AnyRecorder,
    ArtifactRecorder,
    ArtifactRecorderWithGradient,
    RecorderFactory,
    SimpleRecorder,
    SimpleRecorderWithGradient,
)
from ..history.recorder import recorder as _recorder
from . import CallableWithGradient, CostFunction


class Optimizer(ABC):
    """
    Interface for implementing different optimizers.

    Args:
        recorder: recorder object which defines how to store the optimization history.

    """

    def __init__(self, recorder: RecorderFactory = _recorder) -> None:
        self.recorder = recorder

    def minimize(
        self,
        cost_function: CostFunction,
        initial_params: np.ndarray,
        keep_history: bool = False,
    ) -> OptimizeResult:
        """Finds the parameters which minimize given cost function.

        Args:
            cost_function: an object representing the cost function.
            initial_params: initial parameters for the cost function.
            keep_history: flag indicating whether history of cost function
                evaluations should be recorded.
        """
        cost_function = self._preprocess_cost_function(cost_function)
        if keep_history:
            cost_function = self.recorder(cost_function)
        return self._minimize(cost_function, initial_params, keep_history)

    @abstractmethod
    def _minimize(
        self,
        cost_function: CostFunction,
        initial_params: np.ndarray,
        keep_history: bool = False,
    ) -> OptimizeResult:
        """Finds the parameters which minimize given cost function.
        This private method should contain the integration with specific optimizer.

        Args:
            Same as for minimize.
        """
        raise NotImplementedError

    def _preprocess_cost_function(
        self, cost_function: Union[CallableWithGradient, Callable]
    ) -> Union[CallableWithGradient, Callable]:
        """Preprocess cost function before minimizing it.

        This method can be overridden to add some optimizer-specific features
        to cost function. For instance, an optimizer can ensure that the
        cost function has gradient, supplying a default for functions
        without one.

        Args:
            cost_function: a cost function to be preprocessed. Implementers of this
                method shouldn't mutate it.
        Returns:
            preprocess cost function, with the same signature as the original one.
        """
        return cost_function


def optimization_result(
    *, opt_value, opt_params, **kwargs
) -> scipy.optimize.OptimizeResult:
    """Construct instance of OptimizeResult.

    The purpose of this function is to add a safety layer by detecting if required
    components of OptimizationResult are missing already during static analysis.

    Args:
        opt_value: the final value of the function being optimized.
        opt_params: the parameters (arguments) for which opt_value has been achieved.
        kwargs: other attributes (e.g. history) that should be stored in OptimizeResult.
    Returns:
        An instance of OptimizeResult containing opt_value, opt_params and all of the
        other passed arguments.
    """
    return scipy.optimize.OptimizeResult(
        opt_value=opt_value, opt_params=opt_params, **kwargs
    )


def construct_history_info(
    cost_function: CostFunction, keep_history: bool
) -> Dict[str, AnyHistory]:
    histories: Dict[str, AnyHistory] = {}
    if keep_history:
        # TODO: When we upgraded to 3.8 use get_args on AnyRecorder instead
        if isinstance(cost_function, ArtifactRecorder):
            histories["history"] = cost_function.history
        if isinstance(cost_function, SimpleRecorder):
            histories["history"] = cost_function.history
        if isinstance(cost_function, ArtifactRecorderWithGradient):
            histories["gradient_history"] = cost_function.gradient.history
        if isinstance(cost_function, SimpleRecorderWithGradient):
            histories["gradient_history"] = cost_function.gradient.history
    else:
        histories["history"] = cast(AnyHistory, [])
    return histories


def extend_histories(
    cost_function: AnyRecorder, histories: Dict[str, List]
) -> Dict[str, List]:
    new_histories = construct_history_info(cost_function, True)
    updated_histories = {"history": histories["history"] + new_histories["history"]}
    if hasattr(cost_function, "gradient"):
        updated_histories["gradient_history"] = (
            histories["gradient_history"] + new_histories["gradient_history"]
        )
    return updated_histories


class NestedOptimizer(ABC):
    """
    Optimizers that modify cost function throughout optimization.
    An example of such optimizer could be on that freezes certain
    parameters during every iteration or adds new layers of
    the underlying circuit (so called layer-by-layer optimization).

    See MockNestedOptimizer in orquestra.opt.mock_objects for an example.

    Args:
        inner_optimizer: Optimizer object used in the inner optimization loop.
        recorder: recorder object which defines how to store the optimization history.

    Returns:
        An instance of OptimizeResult containing:
            opt_value,
            opt_params,
            nit: total number of iterations of inner_optimizer,
            nfev: total number of calls to cost function,
            history: a list of HistoryEntrys.
                If keep_history is False this should be an empty list.
            gradient_history: if the cost function is a FunctionWithGradient,
                this should be a list of HistoryEntrys representing
                previous calls to the gradient.
    """

    @property
    @abstractmethod
    def inner_optimizer(self) -> Optimizer:
        """Inner optimizer used by this optimizer."""

    @property
    @abstractmethod
    def recorder(self) -> RecorderFactory:
        """Factory for creating recorders of functions being minimized."""
        return _recorder

    def minimize(
        self,
        cost_function_factory: Callable[..., CostFunction],
        initial_params: np.ndarray,
        keep_history: bool = False,
    ) -> OptimizeResult:
        """Finds optimal parameters to minimize the cost function factory.

        Args:
            cost_function_factory: function that generates CostFunction objects.
            initial_params: initial parameters used for optimization
            keep_history: flag indicating whether history of cost function
                evaluations should be recorded.
        """
        return self._minimize(cost_function_factory, initial_params, keep_history)

    @abstractmethod
    def _minimize(
        self,
        cost_function_factory: Callable[..., CostFunction],
        initial_params: np.ndarray,
        keep_history: bool = False,
    ) -> OptimizeResult:
        """Finds the parameters which minimize given cost function factory.
        This private method should contain the integration with specific optimizer.

        Args:
            Same as for minimize.
        """



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""Main implementation of the recorder."""
import copy
from typing import Any, Callable, Dict, Generic, List, NamedTuple, TypeVar, Union

from typing_extensions import overload

from ..api.functions import (
    CallableStoringArtifacts,
    CallableWithGradient,
    CallableWithGradientStoringArtifacts,
    StoreArtifact,
    has_store_artifact_param,
)
from ..api.save_conditions import SaveCondition, always

T = TypeVar("T")
S = TypeVar("S")

NATIVE_RECORDER_ATTRIBUTES = (
    "predicate",
    "history",
    "target",
    "call_number",
    "gradient",
)


class ArtifactCollection(dict):
    """A dict with additional `forced` attribute.

    The `forced` flag is set whenever an artifact is forced into the dictionary
    despite current save_condition being false.
    """

    forced: bool = False


class HistoryEntryWithArtifacts(NamedTuple):
    """A history entry enhanced with artifacts."""

    call_number: int
    params: Any
    value: Any
    artifacts: Dict[str, Any]


class HistoryEntry(NamedTuple):
    """A history entry storing call number, parameters and target function value."""

    call_number: int
    params: Any
    value: Any


def copy_recorder(recorder_to_copy):
    attributes_dict = {
        "target": recorder_to_copy.target,
        "save_condition": recorder_to_copy.predicate,
    }

    recorder_copy = type(recorder_to_copy)(**attributes_dict)
    recorder_copy.call_number = recorder_to_copy.call_number
    recorder_copy.history = recorder_to_copy.history
    return recorder_copy


def deepcopy_recorder(recorder_to_copy, memo):
    attributes_dict = {
        "target": copy.deepcopy(recorder_to_copy.target, memo=memo),
        "save_condition": copy.deepcopy(recorder_to_copy.predicate, memo=memo),
    }

    recorder_copy = type(recorder_to_copy)(**attributes_dict)
    recorder_copy.call_number = recorder_to_copy.call_number
    recorder_copy.history = copy.deepcopy(recorder_to_copy.history, memo=memo)
    return recorder_copy


class SimpleRecorder(Generic[S, T]):
    """A basic recorder that stores history entries.

    Args:
        target: a target function. Calls to the recorder will be propagated to this
          function.
        save_condition: a function determining whether given call should be saved
          to the history. See respective protocol for explanation of this parameter.
    """

    def __init__(self, target: Callable[[S], T], save_condition: SaveCondition):
        self.predicate = save_condition
        self.target = target
        self.history: List[HistoryEntry] = []
        self.call_number = 0

    def __call__(self, params: S) -> T:
        """Call the underlying target function, possibly saving call to the history.

        Args:
            params: argument to be passed to the target function.

        Returns:
            The value returned by the target function.
        """
        return_value = self.target(params)
        if self.predicate(return_value, params, self.call_number):
            self.history.append(
                HistoryEntry(self.call_number, copy.copy(params), return_value)
            )
        self.call_number += 1
        return return_value

    def __getattr__(self, item):
        return getattr(self.target, item)

    def __setattr__(self, key, value):
        if key in NATIVE_RECORDER_ATTRIBUTES:
            return object.__setattr__(self, key, value)
        return setattr(self.target, key, value)

    __copy__ = copy_recorder

    __deepcopy__ = deepcopy_recorder


class SimpleRecorderWithGradient(SimpleRecorder):
    """A recorder saving history entries that works with callables with gradient.

    Except having `gradient` attribute, this recorder is the same as `SimpleRecorder`.
    """

    def __init__(self, target: CallableWithGradient, save_condition: SaveCondition):
        super().__init__(target, save_condition)
        self.gradient = recorder(target.gradient, save_condition)


class ArtifactRecorder(Generic[S, T]):
    """A recorder saving history entries with artifacts.

    Parameters to initializer are the same as for `SimpleRecorder`,
    except the target function should now be capable of storing artifacts.
    """

    def __init__(
        self, target: CallableStoringArtifacts[S, T], save_condition: SaveCondition
    ):
        self.predicate = save_condition
        self.target = target
        self.history: List[HistoryEntryWithArtifacts] = []
        self.call_number = 0

    def __call__(self, params: S) -> T:
        artifacts = ArtifactCollection()
        return_value = self.target(params, store_artifact=store_artifact(artifacts))

        if self.predicate(return_value, params, self.call_number) or artifacts.forced:
            self.history.append(
                HistoryEntryWithArtifacts(
                    self.call_number, copy.copy(params), return_value, artifacts
                )
            )
        self.call_number += 1
        return return_value

    def __getattr__(self, name):
        return getattr(self.target, name)

    def __setattr__(self, name, value):
        if name in ("predicate", "history", "target", "call_number", "gradient"):
            return object.__setattr__(self, name, value)
        return setattr(self.target, name, value)

    __copy__ = copy_recorder

    __deepcopy__ = deepcopy_recorder


class ArtifactRecorderWithGradient(ArtifactRecorder):
    """A recorder storing history entries with artifacts supporting callables with
    gradient.
    """

    def __init__(
        self,
        target: CallableWithGradientStoringArtifacts,
        save_condition: SaveCondition,
    ):
        super().__init__(target, save_condition)
        self.gradient = recorder(target.gradient, save_condition)


def store_artifact(artifacts) -> StoreArtifact:
    """Create a function storing artifacts in given artifacts collection.

    Args:
        artifacts: artifact collection.

    Returns:
        A function with signature:
        _store(artifact_name: str, artifact: Any, force: bool = False) -> None:
        This function is intended to be passed to functions that are capable of
        storing artifacts.
    """

    def _store(artifact_name: str, artifact: Any, force: bool = False) -> None:
        artifacts[artifact_name] = artifact
        if force:
            artifacts.forced = True

    return _store


AnyRecorder = Union[SimpleRecorder, ArtifactRecorder]
RecorderFactory = Callable[[Callable], AnyRecorder]
AnyHistory = Union[List[HistoryEntry], List[HistoryEntryWithArtifacts]]


@overload
def recorder(
    function: CallableWithGradientStoringArtifacts,
    save_condition: SaveCondition = always,
) -> ArtifactRecorderWithGradient:
    """The recorder function: variant for artifact-storing callables with gradient."""


@overload
def recorder(
    function: CallableStoringArtifacts[S, T], save_condition: SaveCondition = always
) -> ArtifactRecorder[S, T]:
    """The recorder function: variant for callables with no gradient that store
    artifacts."""


@overload
def recorder(
    function: CallableWithGradient, save_condition: SaveCondition = always
) -> SimpleRecorderWithGradient:
    """The recorder function: variant for callables with gradient that don't store
    artifacts."""


@overload
def recorder(
    function: Callable[[S], T], save_condition: SaveCondition = always
) -> SimpleRecorder[S, T]:
    """The recorder function: variant for callables without gradient that don't store
    artifacts."""


def recorder(function, save_condition: SaveCondition = always):
    """Create a recorder that is suitable for recording calls to given callable.

    Args:
        function: a callable to be recorded.
        save_condition: a condition on which the calls will be saved. See
          `SaveCondition` protocol for explanation of this parameter. By default
          all calls are saved.

    Returns:
        A callable object (the recorder) wrapping the `function`.
        The return type depends on the passed callable. See overloads defined
        above to check for available variants. Here is a summary:
        - recorder is always callable
        - if `function` has gradient, so does the recorder. Calls to gradient
          and calls made by gradient are NOT recorded.
        - if `function` has possibility to store artifacts (i.e. accepts
          `store_artifact` argument, then so does the recorder.
    """
    with_artifacts = has_store_artifact_param(function)
    with_gradient = isinstance(function, CallableWithGradient)

    if with_artifacts and with_gradient:
        return ArtifactRecorderWithGradient(function, save_condition)
    elif with_artifacts:
        return ArtifactRecorder(function, save_condition)
    elif with_gradient:
        return SimpleRecorderWithGradient(function, save_condition)
    else:
        return SimpleRecorder(function, save_condition)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import dimod
import numpy as np
import pytest
from orquestra.quantum.openfermion import IsingOperator

from orquestra.opt.bqm.conversions import (
    convert_measurements_to_sampleset,
    convert_openfermion_ising_to_qubo,
    convert_qubo_to_openfermion_ising,
    convert_sampleset_to_measurements,
)

from orquestra.quantum.measurements import Measurements  # isort: skip


def test_qubo_conversion_with_binary_fractions():
    qubo = dimod.BinaryQuadraticModel(
        {0: 1, 1: 2, 2: 3},
        {(1, 2): 0.5, (1, 0): -0.25, (0, 2): 2.125},
        -1,
        vartype=dimod.BINARY,
    )
    ising = convert_qubo_to_openfermion_ising(qubo)
    new_qubo = convert_openfermion_ising_to_qubo(ising)
    assert qubo == new_qubo


def test_qubo_conversion_with_non_binary_fractions():
    qubo = dimod.BinaryQuadraticModel(
        {0: 1.01, 1: -2.03, 2: 3},
        {(1, 2): 0.51, (1, 0): -0.9, (0, 2): 2.125},
        -1,
        vartype=dimod.BINARY,
    )
    ising = convert_qubo_to_openfermion_ising(qubo)
    new_qubo = convert_openfermion_ising_to_qubo(ising)

    assert len(qubo.linear) == len(new_qubo.linear)
    assert len(qubo.quadratic) == len(new_qubo.quadratic)

    assert np.isclose(qubo.offset, new_qubo.offset)
    assert qubo.vartype == new_qubo.vartype

    for key in qubo.linear.keys():
        assert np.isclose(qubo.linear[key], new_qubo.linear[key])

    for key in qubo.quadratic.keys():
        assert np.isclose(qubo.quadratic[key], new_qubo.quadratic[key])


def test_converted_ising_evaluates_to_the_same_energy_as_original_qubo():
    qubo = dimod.BinaryQuadraticModel(
        {0: 1, 1: 2, 2: 3},
        {
            (0, 1): 1,
            (0, 2): 0.5,
            (1, 2): 0.5,
        },
        -1,
        vartype=dimod.BINARY,
    )
    all_solutions = [
        [0, 0, 0],
        [0, 0, 1],
        [0, 1, 0],
        [0, 1, 1],
        [1, 0, 0],
        [1, 0, 1],
        [1, 1, 0],
        [1, 1, 1],
    ]

    ising = convert_qubo_to_openfermion_ising(qubo)
    for solution in all_solutions:
        qubo_energy = qubo.energy(solution)
        ising_energy = np.sum(
            Measurements([solution]).get_expectation_values(ising).values
        )
        assert qubo_energy == ising_energy


def test_converted_qubo_evaluates_to_the_same_energy_as_original_ising():
    ising = IsingOperator(
        "2.5 [] + [Z0] + 2[Z0 Z1] +0.5[Z0 Z2] +[Z1] + 0.75[Z1 Z2] -[Z2]"
    )

    all_solutions = [
        [0, 0, 0],
        [0, 0, 1],
        [0, 1, 0],
        [0, 1, 1],
        [1, 0, 0],
        [1, 0, 1],
        [1, 1, 0],
        [1, 1, 1],
    ]

    qubo = convert_openfermion_ising_to_qubo(ising)
    for solution in all_solutions:
        qubo_energy = qubo.energy(solution)
        ising_energy = np.sum(
            Measurements([solution]).get_expectation_values(ising).values
        )
        assert qubo_energy == ising_energy


def test_convert_sampleset_to_measurements():
    bitstrings = [
        (0, 0, 0),
        (0, 0, 1),
        (0, 1, 0),
        (0, 1, 1),
        (1, 0, 0),
        (1, 1, 0),
        (1, 1, 1),
        (1, 0, 1),
        (0, 0, 1),
    ]
    energies = [0 for i in range(len(bitstrings))]
    sampleset = dimod.SampleSet.from_samples(bitstrings, dimod.BINARY, energies)
    target_measurements = Measurements(bitstrings)
    converted_measurements = convert_sampleset_to_measurements(sampleset)

    assert converted_measurements.bitstrings == target_measurements.bitstrings


def test_convert_sampleset_to_measurements_with_change_bitstring_convention():
    bitstrings = [
        (0, 0, 0),
        (0, 0, 1),
        (0, 1, 0),
        (0, 1, 1),
        (1, 0, 0),
        (1, 1, 0),
        (1, 1, 1),
        (1, 0, 1),
        (0, 0, 1),
    ]

    target_bitstrings = [
        (1, 1, 1),
        (1, 1, 0),
        (1, 0, 1),
        (1, 0, 0),
        (0, 1, 1),
        (0, 0, 1),
        (0, 0, 0),
        (0, 1, 0),
        (1, 1, 0),
    ]
    change_bitstring_convention = True
    energies = [0 for i in range(len(bitstrings))]
    sampleset = dimod.SampleSet.from_samples(bitstrings, dimod.BINARY, energies)
    target_measurements = Measurements(target_bitstrings)
    converted_measurements = convert_sampleset_to_measurements(
        sampleset, change_bitstring_convention=change_bitstring_convention
    )

    assert converted_measurements.bitstrings == target_measurements.bitstrings


def test_convert_sampleset_to_measurements_fails_for_non_binary_vartype():
    bitstrings = [
        (0, 0, 0),
    ]
    energies = [0 for _ in range(len(bitstrings))]
    sampleset = dimod.SampleSet.from_samples(bitstrings, dimod.SPIN, energies)
    with pytest.raises(Exception):
        _ = convert_sampleset_to_measurements(sampleset)


def test_convert_sampleset_to_measurements_fails_for_non_int_variables():
    bitstrings = [
        (0, 0, 0),
    ]
    energies = [0 for _ in range(len(bitstrings))]
    sampleset = dimod.SampleSet.from_samples(bitstrings, dimod.SPIN, energies)
    sampleset = sampleset.relabel_variables({0: 0.0, 1: 0.1, 2: 0.2})
    with pytest.raises(Exception):
        _ = convert_sampleset_to_measurements(sampleset)


def test_convert_sampleset_to_measurements_fails_for_variables_from_wrong_range():
    bitstrings = [
        (0, 0, 0),
    ]
    energies = [0 for _ in range(len(bitstrings))]
    sampleset = dimod.SampleSet.from_samples(bitstrings, dimod.SPIN, energies)
    sampleset = sampleset.relabel_variables({0: 1, 1: 2, 2: 3})
    with pytest.raises(Exception):
        _ = convert_sampleset_to_measurements(sampleset)


def test_convert_measurements_to_sampleset_without_qubo():
    bitstrings = [
        (0, 0, 0),
        (0, 0, 1),
        (0, 1, 0),
        (0, 1, 1),
        (1, 0, 0),
        (1, 1, 0),
        (1, 1, 1),
        (1, 0, 1),
        (0, 0, 1),
    ]
    measurements = Measurements(bitstrings)

    target_sampleset = dimod.SampleSet.from_samples(
        bitstrings, dimod.BINARY, [np.nan for _ in bitstrings]
    )
    converted_sampleset = convert_measurements_to_sampleset(measurements)

    # Since energies should be np.nans, using "==" will result in error
    for (target_record, converted_record) in zip(
        target_sampleset.record, converted_sampleset.record
    ):
        for target_element, converted_element in zip(target_record, converted_record):
            np.testing.assert_equal(target_element, converted_element)

    assert converted_sampleset.vartype == target_sampleset.vartype


def test_convert_measurements_to_sampleset_with_change_bitstring_convention():
    bitstrings = [
        (0, 0, 0),
        (0, 0, 1),
        (0, 1, 0),
        (0, 1, 1),
        (1, 0, 0),
        (1, 1, 0),
        (1, 1, 1),
        (1, 0, 1),
        (0, 0, 1),
    ]
    target_bitstrings = [
        (1, 1, 1),
        (1, 1, 0),
        (1, 0, 1),
        (1, 0, 0),
        (0, 1, 1),
        (0, 0, 1),
        (0, 0, 0),
        (0, 1, 0),
        (1, 1, 0),
    ]
    change_bitstring_convention = True
    measurements = Measurements(bitstrings)
    target_sampleset = dimod.SampleSet.from_samples(
        target_bitstrings, dimod.BINARY, [np.nan for _ in bitstrings]
    )
    converted_sampleset = convert_measurements_to_sampleset(
        measurements, change_bitstring_convention=change_bitstring_convention
    )

    # Since energies should be np.nans, using "==" will result in error
    for (target_record, converted_record) in zip(
        target_sampleset.record, converted_sampleset.record
    ):
        for target_element, converted_element in zip(target_record, converted_record):
            np.testing.assert_equal(target_element, converted_element)

    assert converted_sampleset.vartype == target_sampleset.vartype


def test_convert_measurements_to_sampleset_with_qubo():
    bitstrings = [
        (0, 0, 0),
        (0, 0, 1),
        (0, 1, 0),
        (0, 1, 1),
        (1, 0, 0),
        (1, 1, 0),
        (1, 1, 1),
        (1, 0, 1),
        (0, 0, 1),
    ]
    qubo = dimod.BinaryQuadraticModel(
        {0: 1, 1: 2, 2: 3},
        {(1, 2): 0.5, (1, 0): -0.25, (0, 2): 2.125},
        0,
        vartype=dimod.BINARY,
    )
    energies = [0, 3, 2, 5.5, 1, 2.75, 8.375, 6.125, 3]
    measurements = Measurements(bitstrings)

    target_sampleset = dimod.SampleSet.from_samples(bitstrings, dimod.BINARY, energies)
    converted_sampleset = convert_measurements_to_sampleset(measurements, qubo)
    assert target_sampleset == converted_sampleset



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from abc import ABC, abstractmethod
from typing import List, Tuple

import networkx as nx
from orquestra.quantum.openfermion import QubitOperator

from ._problem_evaluation import (
    evaluate_solution,
    solve_graph_problem_by_exhaustive_search,
)


class Problem(ABC):
    @abstractmethod
    def _build_hamiltonian(self, graph: nx.Graph) -> QubitOperator:
        """
        This abstract method is implemented by the subclasses, and
        its goal is to encode the graph of the problem in the form
        of qubit operator.
        """

    def get_hamiltonian(
        self, graph: nx.Graph, scale_factor: float = 1.0, offset: float = 0.0
    ) -> QubitOperator:
        # Relabeling for monotonicity purposes
        num_nodes = range(len(graph.nodes))
        mapping = {node: new_label for node, new_label in zip(graph.nodes, num_nodes)}
        graph = nx.relabel_nodes(graph, mapping=mapping)

        hamiltonian = self._build_hamiltonian(graph)

        hamiltonian.compress()

        return hamiltonian * scale_factor + offset

    def evaluate_solution(self, solution: Tuple[int], graph: nx.Graph) -> float:
        return evaluate_solution(solution, graph, self.get_hamiltonian)

    def solve_by_exhaustive_search(
        self,
        graph: nx.Graph,
    ) -> Tuple[float, List[Tuple[int, ...]]]:
        return solve_graph_problem_by_exhaustive_search(
            graph, cost_function=self.evaluate_solution
        )



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from typing import Callable, Dict, Optional, Sequence, Tuple, Union

import numpy as np
import scipy

from ..api import (
    CallableWithGradient,
    Optimizer,
    construct_history_info,
    optimization_result,
)
from ..history.recorder import RecorderFactory
from ..history.recorder import recorder as _recorder


class ScipyOptimizer(Optimizer):
    def __init__(
        self,
        method: str,
        constraints: Optional[Tuple[Dict[str, Callable]]] = None,
        bounds: Union[
            scipy.optimize.Bounds,
            Sequence[Tuple[float, float]],
            None,
        ] = None,
        options: Optional[Dict] = None,
        recorder: RecorderFactory = _recorder,
    ):
        """
        Integration with scipy optimizers. Documentation for this module is minimal,
        please refer to https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html

        Args:
            method: defines the optimization method
            constraints: list of constraints in the scipy compatible format.
            bounds: bounds for the parameters in the scipy compatible format.
            options: dictionary with additional options for the optimizer.
            recorder: recorder object which defines how to store
                the optimization history.

        """  # noqa: E501
        super().__init__(recorder=recorder)
        self.method = method
        if options is None:
            options = {}
        self.options = options
        self.constraints = [] if constraints is None else constraints
        self.bounds = bounds

    def _minimize(
        self,
        cost_function: Union[CallableWithGradient, Callable],
        initial_params: np.ndarray = None,
        keep_history: bool = False,
    ):
        """
        Minimizes given cost function using functions from scipy.minimize.

        Args:
            cost_function: python method which takes numpy.ndarray as input
            initial_params: initial parameters to be used for optimization
            keep_history: flag indicating whether history of cost function
                evaluations should be recorded.

        """

        jacobian = None
        if isinstance(cost_function, CallableWithGradient) and callable(
            getattr(cost_function, "gradient")
        ):
            jacobian = cost_function.gradient

        result = scipy.optimize.minimize(
            cost_function,
            initial_params,
            method=self.method,
            options=self.options,
            constraints=self.constraints,
            bounds=self.bounds,
            jac=jacobian,
        )
        opt_value = result.fun
        opt_params = result.x

        nit = result.get("nit", None)
        nfev = result.get("nfev", None)

        return optimization_result(
            opt_value=opt_value,
            opt_params=opt_params,
            nit=nit,
            nfev=nfev,
            **construct_history_info(cost_function, keep_history)  # type: ignore
        )



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import numpy as np
import pytest

from orquestra.opt.api import FunctionWithGradient
from orquestra.opt.api.optimizer_test import (
    MANDATORY_OPTIMIZATION_RESULT_FIELDS,
    OptimizerTests,
)
from orquestra.opt.gradients import finite_differences_gradient
from orquestra.opt.optimizers.qiskit_optimizer import QiskitOptimizer


@pytest.fixture(
    params=[
        {"method": "ADAM"},
        {
            "method": "SPSA",
            "optimizer_kwargs": {
                "maxiter": int(1e5),
                "learning_rate": 1e-3,
                "perturbation": 1e-5,
            },
        },
        {
            "method": "AMSGRAD",
            "optimizer_kwargs": {"maxiter": 2e5, "tol": 1e-9, "lr": 1e-4},
        },
    ]
)
def optimizer(request):
    return QiskitOptimizer(**request.param)


# A fast, inaccurate optimizer used when resulting values aren't being tested
@pytest.fixture(
    params=[
        {
            "method": "ADAM",
            "optimizer_kwargs": {
                "maxiter": int(5),
            },
        },
        {
            "method": "SPSA",
            "optimizer_kwargs": {"maxiter": int(5)},
        },
        {
            "method": "AMSGRAD",
            "optimizer_kwargs": {"maxiter": 5},
        },
    ]
)
def low_iter_optimizer(request):
    return QiskitOptimizer(**request.param)


@pytest.fixture(
    params=[
        {
            "method": "NFT",
            "optimizer_kwargs": {"maxiter": 2e5},
        },
    ]
)
def sinusoidal_optimizer(request):
    return QiskitOptimizer(**request.param)


@pytest.fixture(
    params=[
        {
            "method": "NFT",
            "optimizer_kwargs": {"maxiter": 5},
        },
    ]
)
def low_iter_sinusoidal_optimizer(request):
    return QiskitOptimizer(**request.param)


@pytest.fixture(params=[True, False])
def keep_history(request):
    return request.param


class TestQiskitOptimizerTests(OptimizerTests):
    def test_optimizer_succeeds_on_cost_function_without_gradient(
        self, optimizer, sum_x_squared
    ):
        cost_function = sum_x_squared

        results = optimizer.minimize(cost_function, initial_params=np.array([1, -1]))
        assert results.opt_value == pytest.approx(0, abs=1e-5)
        assert results.opt_params == pytest.approx(np.zeros(2), abs=1e-4)

        assert "nfev" in results
        assert "nit" in results
        assert "opt_value" in results
        assert "opt_params" in results
        assert "history" in results

    def test_optimizer_records_history_if_keep_history_is_true(
        self, low_iter_optimizer, sum_x_squared
    ):
        super().test_optimizer_records_history_if_keep_history_is_true(
            low_iter_optimizer, sum_x_squared
        )

    def test_optimizer_does_not_record_history_if_keep_history_is_set_to_false(
        self, low_iter_optimizer, sum_x_squared
    ):
        super().test_optimizer_does_not_record_history_if_keep_history_is_set_to_false(
            low_iter_optimizer, sum_x_squared
        )

    def test_optimizer_does_not_record_history_by_default(
        self, low_iter_optimizer, sum_x_squared
    ):
        super().test_optimizer_does_not_record_history_by_default(
            low_iter_optimizer, sum_x_squared
        )


class TestQiskitSinusoidalOptimizerTests(OptimizerTests):
    """Some optimizers assume that the cost function can be expressed as a sum of
    products of sines and cosines (Eq. 12 of arXiv:1903.12166), a functional form which
    reflects the cost function of many variational quantum algorithms. Such optimizers
    may fail to find the minimum of cost functions that are not of this form.

    This class removes the tests from the base class that check whether the optimizer
    finds the minimum of such non-sinusoidal cost functions, and replaces them with a
    test on a sinuisoidal cost function."""

    @pytest.fixture
    def sum_cos_x(self):
        """Create a sinuisoidal function whose minimum is at the origin with a value of
        zero."""
        return lambda x: -np.sum(np.cos(x)) + x.shape[0]

    @pytest.mark.skip(reason="Optimizer only supports sinusoidal cost functions")
    def test_optimizer_succeeds_with_optimizing_rosenbrock_function(self):
        pass

    @pytest.mark.skip(reason="Optimizer only supports sinusoidal cost functions")
    def test_optimizer_succeeds_with_optimizing_sum_of_squares_function(self):
        pass

    @pytest.mark.skip(reason="Optimizer only supports sinusoidal cost functions")
    def test_optimizer_succeeds_on_cost_function_without_gradient(self):
        pass

    def test_optimizer_succeeds_on_sum_of_cosines_function(
        self, sinusoidal_optimizer, sum_cos_x, keep_history
    ):

        cost_function = FunctionWithGradient(
            sum_cos_x, finite_differences_gradient(sum_cos_x)
        )

        results = sinusoidal_optimizer.minimize(
            cost_function, initial_params=np.array([1, -1]), keep_history=keep_history
        )

        assert results.opt_value == pytest.approx(0, abs=1e-5)
        assert results.opt_params == pytest.approx(np.zeros(2), abs=1e-4)

        assert all(field in results for field in MANDATORY_OPTIMIZATION_RESULT_FIELDS)

        assert "history" in results or not keep_history
        assert "gradient_history" in results or not keep_history

    def test_optimizer_records_history_if_keep_history_is_true(
        self, low_iter_sinusoidal_optimizer, sum_x_squared
    ):
        super().test_optimizer_records_history_if_keep_history_is_true(
            low_iter_sinusoidal_optimizer, sum_x_squared
        )

    def test_optimizer_does_not_record_history_if_keep_history_is_set_to_false(
        self, low_iter_sinusoidal_optimizer, sum_x_squared
    ):
        super().test_optimizer_does_not_record_history_if_keep_history_is_set_to_false(
            low_iter_sinusoidal_optimizer, sum_x_squared
        )

    def test_optimizer_does_not_record_history_by_default(
        self, low_iter_sinusoidal_optimizer, sum_x_squared
    ):
        super().test_optimizer_does_not_record_history_by_default(
            low_iter_sinusoidal_optimizer, sum_x_squared
        )



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import pytest
from orquestra.quantum.openfermion import QubitOperator, change_operator_type
from orquestra.quantum.openfermion.utils import count_qubits

from orquestra.opt.problems import MaxCut, get_random_ising_hamiltonian


class TestGetRandomIsingHamiltonian:
    @pytest.mark.parametrize("num_terms", [2, 6])
    @pytest.mark.parametrize("num_qubits", [2, 5, 7])
    def test_num_qubits_and_num_terms_is_correct(self, num_qubits, num_terms):
        # Given
        if num_qubits >= 5:
            max_number_of_qubits_per_term = 4
        else:
            max_number_of_qubits_per_term = num_qubits

        # When
        hamiltonian = get_random_ising_hamiltonian(
            num_qubits, num_terms, max_number_of_qubits_per_term
        )

        # Then
        # Some qubits may not be included due to randomness, thus the generated number
        # of qubits may be less than `num_qubits`
        assert (
            count_qubits(change_operator_type(hamiltonian, QubitOperator)) <= num_qubits
        )
        generated_num_terms = len(hamiltonian.terms) - 1

        # If two of the randomly generated terms have the same qubits that are operated
        # on, then the two terms will be combined. Therefore, the generated number of
        # terms may be less than `num_terms`
        assert generated_num_terms <= num_terms

    @pytest.mark.parametrize("max_num_terms_per_qubit", [2, 4])
    def test_random_hamiltonian_max_num_terms_per_qubit(self, max_num_terms_per_qubit):
        # Given
        num_qubits = 5
        num_terms = 3

        # When
        hamiltonian = get_random_ising_hamiltonian(
            num_qubits, num_terms, max_num_terms_per_qubit
        )

        # Then
        for term in hamiltonian.terms.keys():
            assert len(term) <= max_num_terms_per_qubit



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import copy
from typing import Callable, Optional, Union

import numpy as np
from scipy.optimize import OptimizeResult

from ..api import (
    CallableWithGradient,
    Optimizer,
    construct_history_info,
    optimization_result,
)
from ..history.recorder import RecorderFactory
from ..history.recorder import recorder as _recorder


class SimpleGradientDescent(Optimizer):
    def __init__(
        self,
        learning_rate: float,
        number_of_iterations: int,
        recorder: RecorderFactory = _recorder,
    ):
        """
        Args:
            parameter_values_list: list of parameter values to evaluate
            recorder: recorder object which defines how to store
                the optimization history.
        """
        super().__init__(recorder=recorder)
        self.learning_rate = learning_rate

        assert number_of_iterations > 0
        self.number_of_iterations = number_of_iterations

    def _minimize(
        self,
        cost_function: Union[CallableWithGradient, Callable],
        initial_params: np.ndarray,
        keep_history: bool = False,
    ) -> OptimizeResult:
        """
        Finds the parameters which minimize given cost function, by trying all
        the parameters from the provided list of points.

        Note:
            This optimizer does not require evaluation of the cost function,
            but relies only on gradient evaluation. This means, that if we want to
            keep track of values of the cost functions for each iteration, we
            need to perform extra evaluations. Therefore using `keep_history=True`
            will add extra evaluations that are not necessary for
            the optimization process itself.

        Args:
            cost_function: object representing cost function we want to minimize
            inital_params: initial parameters for the cost function
            keep_history: flag indicating whether history of cost function
                evaluations should be recorded. Using this will increase runtime,
                see note.

        """
        assert isinstance(cost_function, CallableWithGradient)

        current_parameters = copy.deepcopy(initial_params)
        for _ in range(self.number_of_iterations):
            gradients = cost_function.gradient(current_parameters)
            current_parameters = current_parameters - (self.learning_rate * gradients)
            if keep_history:
                final_value = cost_function(current_parameters)

        if not keep_history:
            final_value = cost_function(current_parameters)

        return optimization_result(
            opt_value=final_value,
            opt_params=current_parameters,
            nit=self.number_of_iterations,
            nfev=None,
            **construct_history_info(cost_function, keep_history),  # type: ignore
        )



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""Examples of functions for which recorder works.

Note that none of this use ValueEstimate as a return value. This is because
recorders can work with callable objects of any return type.
"""
from typing import Optional

import numpy as np

from .functions import (
    FunctionWithGradient,
    FunctionWithGradientStoringArtifacts,
    StoreArtifact,
)


def sum_of_squares(params):
    """Multidimensional sum of squares.

    Args:
        params: numbers to be squared and summed.

    Returns:
        Sum of squares of numbers in params.
    """
    return sum(x**2 for x in params)


# The below implements function: f(x, y) = x ** 2  - x * y
function_1 = FunctionWithGradient(
    function=lambda params: params[0] ** 2 - params[0] * params[1],
    gradient=lambda params: np.array([2 * params[0] - params[1], -params[0]]),
)


class Function2:
    """Example of function with gradient, this time implemented as a class.

    This function is f((x, y, z)) = multiplier * (x - xy + z ** 2),
    where multiplier is fixed.

    Args:
        multiplier: fixed multiplier, see equation above.
    """

    def __init__(self, multiplier):
        self.multiplier = multiplier

    def __eq__(self, other):
        return type(self) == type(other) and self.multiplier == other.multiplier

    def __call__(self, params):
        """Compute value of this function.

        Args:
            params: parameters of the function. Their expected length is 3.

        Returns:
            self.multiplier * (x - xy + z ** 2) where x, y, z = params.
        """
        return self.multiplier * (params[0] - params[0] * params[1] + params[2] ** 2)

    def gradient(self, params):
        """Compute value of gradient of this function for given params.

        Args:
            params: parameters of the gradient. The expected length is 3.

        Returns:
            self.multiplier * [1 - y, -x, 2z] where x, y, z = params.
        """
        return self.multiplier * np.array([1 - params[1], -params[0], 2 * params[2]])


def function_3(params: int, store_artifact=None):
    """An example of function that stores artifacts.

    Args:
        params: parameters for the function. Note that the name `params` is
          to follow the general conversion, but the function expects
          integer.
        store_artifact: callback for storing artifacts. See StoreArtifact
          protocol for explanation.

    Returns:
        The input argument multiplied by 2. As a side effect, binary
        representation of the input argument is stored as "bitstring"
        artifact.
    """
    bitstring = bin(params)[2:]
    if store_artifact:
        store_artifact("bitstring", bitstring)
    return 2 * params


def function_4(params: int, store_artifact: Optional[StoreArtifact] = None) -> int:
    """Another example of function that stores artifacts, but occasionally forces store.

    This function is identical to function_3, except the artifact is forcefully
    stored if input parameter is divisible by 2.
    """
    bitstring = bin(params)[2:]
    if store_artifact:
        store_artifact("bitstring", bitstring, force=params % 2 == 0)
    return 2 * params


class Function5:
    """Example of a function that stores artifacts and has gradient. Full service.

    This function is f((x, y, z)) = alpha * xyz, where alpha is fixed.

    Args:
        alpha: parameter for this function, see equation above.
    """

    def __init__(self, alpha):
        self.alpha = alpha

    def __eq__(self, other):
        return type(self) == type(other) and self.alpha == other.alpha

    def __call__(self, params: np.ndarray, store_artifact=None):
        """Compute value of this function.

        Args:
            params: parameters of the function. The expected length is 3.

        Returns:
            self.alpha * x * y * z, where x, y, z = params
        """
        if params[0] > 0 and store_artifact:
            store_artifact("something", params[0] + params[1])
        return self.alpha * params[0] * params[1] * params[2]

    def gradient(self, params):
        """Compute value of gradient of this function for given params.

        Args:
            params: parameters of the gradient. The expected length is 3.

        Returns:
            self.alpha * [yz, xz, xy] where x, y, z = params.
        """
        return self.alpha * np.array(
            [params[1] * params[2], params[0] * params[2], params[0] * params[1]]
        )


function_6 = FunctionWithGradientStoringArtifacts(
    function=Function5(5), gradient=Function5(5).gradient
)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import pytest

from orquestra.opt.api.optimizer_test import OptimizerTests
from orquestra.opt.mock_objects import mock_cost_function
from orquestra.opt.optimizers.cma_es_optimizer import CMAESOptimizer


@pytest.fixture(scope="function")
def optimizer():
    return CMAESOptimizer(sigma_0=0.1)


@pytest.fixture(params=[True, False])
def keep_history(request):
    return request.param


class TestCMAESOptimizer(OptimizerTests):
    def test_cmaes_specific_fields(self):
        results = CMAESOptimizer(
            sigma_0=0.1, options={"maxfevals": 99, "popsize": 5}
        ).minimize(mock_cost_function, initial_params=[0, 0], keep_history=True)

        assert "cma_xfavorite" in results
        assert isinstance(results["cma_xfavorite"], list)
        assert len(results["history"]) == 100



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from typing import Optional

import dimod
import numpy as np
from dimod import BinaryQuadraticModel, SampleSet
from orquestra.quantum.measurements import Measurements
from orquestra.quantum.openfermion import IsingOperator


def convert_qubo_to_openfermion_ising(qubo: BinaryQuadraticModel) -> IsingOperator:
    """Converts dimod BinaryQuadraticModel to OpenFermion IsingOperator object.

    The resulting Openfermion IsingOperator has the following property: for every
    bitstring, its expected value is the same as the energy of the original QUBO.
    In order to ensure this, we had to add a minus sign for the coefficients
    of the linear terms coming from dimod conversion.
    For more context about conventions used please refer to note in
    `convert_measurements_to_sampleset` docstring.

    Args:
        qubo: Object we want to convert

    Returns:
        IsingOperator: IsingOperator representation of the input qubo.

    """
    linear_coeffs, quadratic_coeffs, offset = qubo.to_ising()

    list_of_ising_strings = [f"{offset}[]"]

    for i, value in linear_coeffs.items():
        list_of_ising_strings.append(f"{-value}[Z{i}]")

    for (i, j), value in quadratic_coeffs.items():
        list_of_ising_strings.append(f"{value}[Z{i} Z{j}]")

    ising_string = " + ".join(list_of_ising_strings)
    return IsingOperator(ising_string)


def convert_openfermion_ising_to_qubo(operator: IsingOperator) -> BinaryQuadraticModel:
    """
    Converts dimod Openfermion IsingOperator to BinaryQuadraticModel object.
    The resulting QUBO has the following property:
    For every bitstring, its energy is the same as the expected value of the original
    Ising Hamiltonian. For more context about conventions used please refer to note in
    `convert_measurements_to_sampleset` docstring.

    Note:
        The conversion might not be 100% accurate due to performing floating point
        operations during conversion between Ising and QUBO models.

    Args:
        operator: IsingOperator we want to convert
    Returns:
        qubo: BinaryQuadraticModel representation of the input operator

    """

    if not isinstance(operator, IsingOperator):
        raise TypeError(
            f"Input is of type: {type(operator)}. Only Ising Operators are supported."
        )
    offset = 0
    linear_terms = {}
    quadratic_terms = {}
    for term, coeff in operator.terms.items():
        if len(term) == 0:
            offset = coeff
        if len(term) == 1:
            linear_terms[term[0][0]] = -coeff
        if len(term) == 2:
            quadratic_terms[(term[0][0], term[1][0])] = coeff
        if len(term) > 2:
            raise ValueError(
                "Ising to QUBO conversion works only for quadratic Ising models."
            )

    dimod_ising = BinaryQuadraticModel(
        linear_terms, quadratic_terms, offset, vartype="SPIN"
    )
    return dimod_ising.change_vartype(dimod.Vartype.BINARY, inplace=False)


def convert_sampleset_to_measurements(
    sampleset: SampleSet,
    change_bitstring_convention: bool = False,
) -> Measurements:
    """
    Converts dimod SampleSet to orquestra.quantum Measurements.
    Works only for the sampleset with "BINARY" vartype and variables being range of
    integers starting from 0.

    Note:
        Since Measurements doesn't hold information about the energy of the samples,
        this conversion is lossy. For more explanation regarding
        change_bitstring_convention please read docs of
        `convert_measurements_to_sampleset`.

    Args:
        sampleset: SampleSet we want to convert
        change_bitstring_convention: whether to flip the bits in bitstrings to, depends
        on the convention one is using (see note).

    """
    if sampleset.vartype != dimod.BINARY:
        raise TypeError("Sampleset needs to have vartype BINARY")
    for i in range(max(sampleset.variables)):
        if sampleset.variables[i] != i:
            raise ValueError(
                "Variables of sampleset need to be ordered list of integers"
            )

    bitstrings = [
        tuple(int(change_bitstring_convention != sample[i]) for i in range(len(sample)))
        for sample in sampleset.samples()
    ]
    return Measurements(bitstrings)


def convert_measurements_to_sampleset(
    measurements: Measurements,
    bqm: Optional[BinaryQuadraticModel] = None,
    change_bitstring_convention: bool = False,
) -> SampleSet:
    """
    Converts dimod SampleSet to orquestra.quantum Measurements.
    If no bqm is specified, the vartype of the SampleSet will be "BINARY" and the
    energies will be NaN. If bqm is specified, its vartype will be preserved and
    the energy values will be calculated.

    Note:
        The convention commonly used in quantum computing is that 0 in a bitstring
        represents eigenvalue 1 of an Ising Hamiltonian,and 1 represents eigenvalue -1.
        However, there is another convention, used in dimod, where the mapping is
        0 -> -1 and 1 -> 1 instead. Therefore if we try to use the bitstrings coming
        from solving the problem framed in one convention to evaluate energy for
        problem state in the second one, the results will be incorrect. This can be
        fixed with changing the value of `change_bitstring_convention` flag. Currently
        we changed the conventions appropriately in `convert_qubo_to_openfermion_ising`
        and `convert_openfermion_ising_to_qubo`, however, this still might be an issue
        in cases where qubo/Ising representation is created using different tools.
        It's hard to envision a specific example at the time of writing, but experience
        shows that it needs to be handled with caution.
    Args:
        measurements: Measurements object to be converted
        bqm: if provided, SampleSet will include energy values for each sample.
        change_bitstring_convention: whether to flip the bits in bitstrings to, depends
            on the convention one is using (see note).
    Returns:
        SampleSet object
    """
    bitstrings = [
        tuple(int(change_bitstring_convention != bit) for bit in bitstring)
        for bitstring in measurements.bitstrings
    ]

    if not bqm:
        return SampleSet.from_samples(
            bitstrings, "BINARY", [np.nan for _ in bitstrings]
        )
    if bqm.vartype != dimod.BINARY:
        raise TypeError("BQM needs to have vartype BINARY")

    return SampleSet.from_samples_bqm(bitstrings, bqm)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""Implementation of cost function augmentations.

The "augmentation" is a function returning function. Augmentations can add side effects
to the function invocation (like logging or recording history), or even modify the
return value.

Cost functions are augmented via `augment_cost_function`. Basic usage looks as follows

augmented_func = augment_cost_function(func, [function_logger(level=level)])

If the function to be augmented has gradient, separate augmentations can be applied
to the function and the gradient, e.g.:

augmented_func = augment_cost_function(
    func,
    cost_function_augmentations=[function_logger(level=logging.INFO)],
    gradient_augmentations=[function_logger(level=logging.DEBUG)]
)


In principle, any function mapping cost function to cost function can be used as
augmentation. The common pattern however is an augmentation that triggers a side
effect, possibly when some conditions are met, and otherwise leaves the augmented
function unchanged. The `function_logger` augmentations is an example of such
augmentation.

There is a shortcut for implementing augmentations like function_logger. The required
steps are:
- Create class inheriting ConditionalSideEffect.
- Implement _act method, that defines what the side effect is.
- Create a thin wrapper function that constructs actual augmentation.
Refer to `function_logger` below for an example of this process.
"""
import abc
import logging
from typing import Callable, Iterable, Optional

from .api import (
    CallableStoringArtifacts,
    CallableWithGradient,
    CallableWithGradientStoringArtifacts,
    CostFunction,
    SaveCondition,
    always,
)

FunctionAugmentation = Callable[[Callable], Callable]


class _AttributePropagatingWrapper:
    """A wrapper for handling attribute propagation.

    This is intended to for handling attribute propagation between an object
    implementing augmentation and the function being augmented.

    This wrapper propagates all attribute accesses and calls to the wrapped `function`.
    If the attribute is not present on the `function`, `additional_source_of_attributes`
    is tried. If that fails, usual AttributeError is raised.

    Attributes can be overrides via attribute_overrides. This is to ensure that some
    attributes of wrapper objects are accessible even if they are also defined in the
    wrapped objects.
    """

    def __init__(self, function, additional_attribute_source, attribute_overrides):
        # We use object.__setattr__ here to not trigger propagation mechanism
        # implemented below. In  particular, usual self._function = function would fail
        # because it would already try to find the attribute on self._function, which
        # is undefined.
        object.__setattr__(self, "_function", function)
        object.__setattr__(
            self, "_additional_attribute_source", additional_attribute_source
        )
        object.__setattr__(
            self,
            "_attribute_overrides",
            attribute_overrides if attribute_overrides else {},
        )

    def __call__(self, *args, **kwargs):
        """Propagate call to the wrapped function.

        This is needed for this object to be callable.
        """
        return self._function(*args, **kwargs)

    def __getattr__(self, attr_name):
        """Get attribute dynamically.

        This dunder method is called only if attr_name attribute is not
        present on the object itself. This implementation tries to find
        such an attribute in the following sources (in that order):
        - attribute overrides
        - wrapped function
        - additional attribute sources

        Overriding __getattr__ is crucial for propagating attribute queries to the
        wrapped objects.
        """
        try:
            return self._attribute_overrides[attr_name]
        except KeyError:
            try:
                return getattr(self._function, attr_name)
            except AttributeError:
                return getattr(self._additional_attribute_source, attr_name)

    def __setattr__(self, attr_name, value):
        """Set attribute not defined in this object's attribute dictionary.

        This dunder method is called only if attr_name is not present on the
        object itself. This implementation sets value of this attribute on the
        wrapped function if it is already present here, and in the additional
        attribute source otherwise.

        Overriding __setattr__ is crucial for propagating attribute settings to the
        wrapped objects.
        """
        setattr(
            self._function
            if hasattr(self._function, attr_name)
            else self._additional_attribute_source,
            attr_name,
            value,
        )


def _augment_function(function, augmentations, attribute_overrides):
    """Augment a function, possibly overriding some attributes."""
    if not augmentations:
        return function
    augmented_function = function
    for augmentation in augmentations:
        augmented_function = _AttributePropagatingWrapper(
            augmentation(augmented_function), augmented_function, attribute_overrides
        )

    return augmented_function


def augment_cost_function(
    cost_function: CostFunction,
    cost_function_augmentations: Optional[Iterable[FunctionAugmentation]] = None,
    gradient_augmentations: Optional[Iterable[FunctionAugmentation]] = None,
):
    """Augment a function and its gradient.

    Args:
        cost_function: a function to be augmented.
        cost_function_augmentations: augmentations to be applied to cost_function
            itself, in left-to-right order.
        gradient_augmentations: augmentations to be applied to a gradient. If the
            cost_function has no `gradient` attribute, this argument is ignored.
    Returns:
        A function with all application applied to it. The returned object
        has all attributes present in cost_function. The same is true about
        gradient. The exact behaviour of the returned object depends on the
        augmentations used, and the order in which they were applied.
    """
    attribute_overrides = (
        {
            "gradient": _augment_function(
                cost_function.gradient, gradient_augmentations, {}
            )
        }
        if isinstance(cost_function, CallableWithGradient)
        or isinstance(cost_function, CallableWithGradientStoringArtifacts)
        else {}
    )

    augmented_cost_function = _augment_function(
        cost_function, cost_function_augmentations, attribute_overrides
    )

    return augmented_cost_function


class ConditionalSideEffect(abc.ABC):
    """Base class for implementing augmentations that trigger some side effect.

    ConditionalSideEffect implementations don't modify the return value of
    the wrapped function, but trigger some additional action, possibly
    only when some conditions are met.

    Args:
        function: a single argument function to be wrapped
        predicate: a function with signature (result, params, call_number) where
          function(params) = result and call_number indicates how many times this
          object has been called so far. The side effect is triggerred if and only
          if predicate(result, params, call_number) is True. By default, side
          effect is triggerred always.

    Attributes:
        function: function passed to the initializer
        predicate: predicate passed to the initializer
        call_number: number indicating how many calls have been made to this
          function so far.

    Notes:
        Since no synchronization is implemented, ConditionalSideEffects are inherently
        unsuitable for multi-threaded use.

        Concrete subclasses of ConditionalSideEffect need to implement _act method.
    """

    def __init__(self, function: Callable, predicate: SaveCondition = always):
        self.function = function
        self.predicate = predicate
        self.call_number = 0

    def __call__(self, params):
        result = self.function(params)
        if self.predicate(result, params, self.call_number):
            self._act(result, params)

        self.call_number += 1
        return result

    @abc.abstractmethod
    def _act(self, result, params):
        """Perform side effect defined by this object.

        Args:
            result: value of wrapped function.
            params; corresponding parameters (i.e. self.function(params) = result.
        """


class FunctionLogger(ConditionalSideEffect):
    """Side effect causing function call to be logged.

    Args:
        function: function to be wrapped.
        predicate: boolean function defining when the logging should happen. See
          ConditionalSideEffect for more detailed description.
        logger: logger to be used. If not provided, logger with __name__ name will
          be used.
        level: log level to be used for logging messages. The default is INFO.
        message: %-style template string for message to be printed. Message will be
          constructed by passing format arguments (call number, function value).
    """

    def __init__(
        self,
        function: Callable,
        predicate: SaveCondition = always,
        logger=None,
        level=logging.INFO,
        message="Function called. Call number: %d, function value: %s",
    ):
        super().__init__(function, predicate)
        self.logger = logging.getLogger(__name__) if logger is None else logger
        self.level = level
        self.message = message

    def _act(self, result, params):
        self.logger.log(self.level, self.message, self.call_number, result)


def function_logger(
    predicate: SaveCondition = always,
    logger=None,
    level=logging.INFO,
    message="Function called. Call number: %d, function value: %s",
):
    """Cost function augmentation adding logging.

    For description of parameters see FunctionLogger class above.
    """

    def _augment(function):
        return FunctionLogger(function, predicate, logger, level, message)

    return _augment



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from typing import Callable

import numpy as np
import pytest

from orquestra.opt.api import CostFunction
from orquestra.opt.api.optimizer_test import NESTED_OPTIMIZER_CONTRACTS
from orquestra.opt.mock_objects import (
    MockNestedOptimizer,
    MockOptimizer,
    mock_cost_function,
)


class MaliciousNestedOptimizer(MockNestedOptimizer):
    def _minimize(
        self,
        cost_function_factory: Callable[[int], CostFunction],
        initial_params: np.ndarray,
        keep_history: bool = False,
    ):
        keep_history = not keep_history
        results = super()._minimize(
            cost_function_factory, initial_params, keep_history=keep_history
        )
        del results["nit"]
        return results


_good_nested_optimizer = MockNestedOptimizer(inner_optimizer=MockOptimizer(), n_iters=5)
_malicious_nested_optimizer = MaliciousNestedOptimizer(
    inner_optimizer=MockOptimizer(), n_iters=5
)


def mock_cost_function_factory(iteration_id: int):
    def modified_cost_function(params):
        return mock_cost_function(params) ** iteration_id

    return modified_cost_function


@pytest.mark.parametrize("contract", NESTED_OPTIMIZER_CONTRACTS)
def test_validate_contracts(contract):
    assert contract(
        _good_nested_optimizer,
        mock_cost_function_factory,
        np.array([2]),
    )
    assert not contract(
        _malicious_nested_optimizer,
        mock_cost_function_factory,
        np.array([2]),
    )



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""Module with definitions of gradient."""
import numpy as np


def finite_differences_gradient(function, finite_diff_step_size=1e-5):
    """Create a (central) finite differences gradient for a given function.

    Args:
        function: callable accepting 1-D numpy arrays and returning float.
        finite_diff_step_size: finite difference size used to estimate gradient.
    Returns:
        A function that returns a gradient estimation using central finite
        differences method.
    """

    def _gradient(parameters):
        gradient = np.array([])
        for idx in range(len(parameters)):
            values_plus = parameters.astype(float)
            values_minus = parameters.astype(float)
            values_plus[idx] += finite_diff_step_size
            values_minus[idx] -= finite_diff_step_size
            gradient = np.append(
                gradient,
                (function(values_plus) - function(values_minus))
                / (2 * finite_diff_step_size),
            )
        return gradient

    return _gradient



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import networkx as nx
from orquestra.quantum.openfermion import QubitOperator

from ..api.problem import Problem


class MaxIndependentSet(Problem):
    """Solves maximum independent set problem on an undirected graph using an
    ising model formulation.

    The solution to a maximum independent set is the largest set of nodes
    which do not share an edge.
    Based on "Efficient Combinatorial Optimization Using Quantum Annealing" p. 8
    (https://arxiv.org/pdf/1801.08653.pdf)
    and also mentioned briefly in
    "Ising formulations of many NP problems" by A. Lucas, page 11 section 4.2
    (https://arxiv.org/pdf/1302.5843.pdf).
    """

    def _build_hamiltonian(self, graph: nx.Graph) -> QubitOperator:
        """Construct a qubit operator with Hamiltonian for the maximum independent
        set problem.

        The operator's terms contain Pauli Z matrices applied to qubits. The qubit
        indices are based on graph node indices in the graph definition, not on the
        node names.

        Args:
            graph: undirected weighted graph defining the problem
        """
        ham_a = QubitOperator()
        for i, j in graph.edges:
            ham_a += (1 - QubitOperator(f"Z{i}")) * (1 - QubitOperator(f"Z{j}"))

        ham_b = QubitOperator()
        for i in graph.nodes:
            ham_b += QubitOperator(f"Z{i}")

        return ham_a / 2 + ham_b / 2 - len(graph.nodes) / 2



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""Test cases for storing artifacts and recording functions that store them."""
import numpy as np
import pytest

from orquestra.opt.api import CallableStoringArtifacts
from orquestra.opt.api.example_functions import Function5, function_3, function_4
from orquestra.opt.api.save_conditions import every_nth
from orquestra.opt.history.recorder import ArtifactCollection, recorder, store_artifact


def test_store_artifact_by_default_does_not_force_artifacts_storage():
    artifacts = ArtifactCollection()
    store_artifact(artifacts)("bitstring", "1111")
    assert not artifacts.forced


@pytest.mark.parametrize(
    "source_function,params_sequence,expected_artifacts",
    [
        (
            function_3,
            [3, 4, 5],
            [{"bitstring": string} for string in ["11", "100", "101"]],
        ),
        (
            function_4,
            [0, 10, 21],
            [{"bitstring": string} for string in ["0", "1010", "10101"]],
        ),
        (
            Function5(2),
            [np.array([1.5, 2, 3]), np.array([4.0, 2.5, 3.0])],
            [{"something": 3.5}, {"something": 6.5}],
        ),
    ],
)
def test_recorder_stores_all_artifacts_by_default(
    source_function: CallableStoringArtifacts, params_sequence, expected_artifacts
):
    function = recorder(source_function)
    for param in params_sequence:
        function(param)

    assert [entry.artifacts for entry in function.history] == expected_artifacts


def test_recorder_stores_history_entry_if_artifact_is_force_stored():
    function = recorder(function_4, save_condition=every_nth(5))
    for n in [0, 2, 3, 5, 7, 5]:
        function(n)

    assert [entry.call_number for entry in function.history] == [0, 1, 5]
    assert [entry.value for entry in function.history] == [0, 4, 10]
    assert [entry.params for entry in function.history] == [0, 2, 5]
    assert [entry.artifacts for entry in function.history] == [
        {"bitstring": "0"},
        {"bitstring": "10"},
        {"bitstring": "101"},
    ]



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import networkx as nx
from orquestra.quantum.openfermion import QubitOperator

from ..api.problem import Problem


class VertexCover(Problem):
    """Solves vertex cover problem on an undirected graph using an ising model
    formulation.

    The solution of a vertex cover problem is the minimal number of colored
    verticies such that all edges connect to a colored vertex.
    From "Ising formulations of many NP Problems" by A. Lucas, eq. 33 and 34
    (https://arxiv.org/pdf/1302.5843.pdf)
    and
    https://quantumcomputing.stackexchange.com/questions/16082/vertex-cover-mappings-from-qubo-to-ising-and-vice-versa
    for corrective translation shifts

    Args:
        A: Cost of having an edge which is not connected to a colored vertex.
        Should be large (_A = 5) to ensure output is a valid solution.
        B: Cost of coloring a particular vertex.

    Attributes:
        _A: See argument A above.
        _B: See argument B above.
    """

    def __init__(self, A: int = 5, B: int = 1):
        self._A = A
        self._B = B

    def _build_hamiltonian(
        self,
        graph: nx.Graph,
    ) -> QubitOperator:
        """Construct a Hamiltonian for the vertex cover problem.

        The operator's terms contain Pauli Z matrices applied to qubits. The qubit
        indices are based on graph node indices in the graph definition, not on the
        node names.

        Args:
            graph: undirected weighted graph defining the problem
        """  # noqa: E501
        ham_a = QubitOperator()
        for i, j in graph.edges:
            ham_a += (1 - QubitOperator(f"Z{i}")) * (1 - QubitOperator(f"Z{j}"))
        ham_a *= self._A / 4

        ham_b = QubitOperator()
        for i in graph.nodes:
            ham_b += QubitOperator(f"Z{i}")
        ham_b *= self._B / 2

        return ham_a + ham_b + len(graph.nodes) * self._B / 2



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""Save conditions possible to use with recorder."""
from typing import Any

from typing_extensions import Protocol


class SaveCondition(Protocol):
    """Protocol of a function determining if given call should should be saved in the
    history."""

    def __call__(self, value: Any, params: Any, call_number: int) -> bool:
        """Determine whether current call should be saved in the history.

        Suppose the recorder is constructed for a function `f`, and the params
        `x` are passed to `f` such that `y`=`f(x)`. Then, if this is `n-th`
        evaluation of the function, the value of __call__(y, x, n) determines
        if current call should be saved to the history.

        Args:
            value: current value of the function.
            params: parameters passed to the function.
            call_number: a natural number determining how many times the target
              function has been called.

        Returns:
            A boolean indicating whether the call being processed should be
            saved to history.
        """


def always(value: Any, params: Any, call_number: int) -> bool:
    """Default save condition: save always.

    See parameters in SaveCondition.__call__ for explanation."""
    return True


def every_nth(n: int) -> SaveCondition:
    """Save condition: every n-th step, counting from zero-th one.

    Note: this is factory function, i.e. it returns the actual save condition
    for given n.

    Args:
        n: the integer determining which steps should be saved in history.

    Returns:
        Function `f` implementing the SaveCondition protocol, such that
        f(y, x, k) is True if and only if k = 0 (mod n).
    """

    def _save_condition(value: Any, params: Any, call_number: int) -> bool:
        return call_number % n == 0

    return _save_condition



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""Protocols describing different kinds of functions."""
from inspect import signature
from typing import Any, Callable, NamedTuple, Optional, TypeVar, Union, cast

import numpy as np
from typing_extensions import Protocol, runtime_checkable

T = TypeVar("T", covariant=True)
S = TypeVar("S", contravariant=True)


class StoreArtifact(Protocol):
    """A protocol describing how the artifacts are stored."""

    def __call__(self, artifact_name: str, artifact: Any, force: bool = False) -> None:
        pass


@runtime_checkable
class CallableWithGradient(Protocol):
    """A callable with gradient."""

    def __call__(self, params: np.ndarray) -> float:
        pass

    def gradient(self, params: np.ndarray) -> np.ndarray:
        pass


@runtime_checkable
class CallableStoringArtifacts(Protocol[S, T]):
    """A callable that stores artifacts."""

    def __call__(self, params: S, store_artifact: Optional[StoreArtifact]) -> T:
        pass


@runtime_checkable
class CallableWithGradientStoringArtifacts(
    CallableStoringArtifacts[np.ndarray, T], Protocol
):
    """A callable with gradient that stored artifacts."""

    def gradient(self, params: np.ndarray) -> np.ndarray:
        pass


def has_store_artifact_param(function) -> bool:
    """Determine if given callable is capable of storing artifacts.

    :param function: a callable to be checked.
    :return: True, if `function` has store_artifact parameter and False otherwise.
    """
    try:
        return "store_artifact" in signature(function, follow_wrapped=True).parameters
    except ValueError:
        # Rationale:
        # The only callables that are of interest to us that aren't supported by
        # signature that I am aware of are numpy ufunc's. Obviously, they don't have
        # store_artifact parameter.
        return False


class FunctionWithGradient(NamedTuple):
    """A callable with gradient."""

    function: Callable[[np.ndarray], float]
    gradient: Callable[[np.ndarray], np.ndarray]

    def __call__(self, params: np.ndarray) -> float:
        return self.function(params)

    def __getattr__(self, name):
        return getattr(self.function, name)

    def __setattr__(self, name, value):
        return setattr(self.function, name, value)


class FunctionWithGradientStoringArtifacts(NamedTuple):
    """A callable with gradient that also stores artifacts."""

    function: CallableStoringArtifacts
    gradient: Callable[[np.ndarray], np.ndarray]

    def __call__(
        self, params: np.ndarray, store_artifact: StoreArtifact = None
    ) -> float:
        return self.function(params, store_artifact)

    def __getattr__(self, name):
        return getattr(self.function, name)

    def __setattr__(self, name, value):
        return setattr(self.function, name, value)


def function_with_gradient(
    function: Union[Callable[[np.ndarray], float], CallableStoringArtifacts],
    gradient: Callable[[np.ndarray], np.ndarray],
) -> Union[FunctionWithGradient, FunctionWithGradientStoringArtifacts]:
    """Combine function and gradient into an entity adhering to protocols used by
    history recorder.

    Note that this is a preferred method for adding gradient to your function,
    as it should automatically detect whether the function stores artifact or not.
    """
    if has_store_artifact_param(function):
        return FunctionWithGradientStoringArtifacts(
            cast(CallableStoringArtifacts, function), gradient
        )
    else:
        return FunctionWithGradient(
            cast(Callable[[np.ndarray], float], function), gradient
        )



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import networkx as nx
from orquestra.quantum.openfermion import QubitOperator

from ..api.problem import Problem


class GraphPartitioning(Problem):
    """Solves the graph partitioning problem for undirected, unweighted graph using
    an ising model formulation.

    The solution of a graph partitioning  problem is the set of nodes S that minimizes
    the weight of the edges between S and the rest of the graph. S must contain half
    the nodes in the graph.

    From "Ising formulations of many NP problems" by A. Lucas, page 6, eqs 7-9
    (https://arxiv.org/pdf/1302.5843.pdf).
    """

    def _build_hamiltonian(self, graph: nx.Graph) -> QubitOperator:
        """Construct a qubit operator with Hamiltonian for the graph partition problem.

        The returned Hamiltonian is consistent with the definitions from
        "Ising formulations of many NP problems" by A. Lucas, page 6
        (https://arxiv.org/pdf/1302.5843.pdf).

        The operator's terms contain Pauli Z matrices applied to qubits. The qubit
        indices are based on graph node indices in the graph definition, not on the
        node names.

        Args:
            graph: undirected weighted graph defining the problem
        """
        ham_a = QubitOperator()
        for i in graph.nodes:
            ham_a += QubitOperator(f"Z{i}")
        ham_a = ham_a**2

        ham_b = QubitOperator()
        for i, j in graph.edges:
            ham_b += 1 - QubitOperator(f"Z{i} Z{j}")
        ham_b /= 2

        return ham_a + ham_b



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import pytest
from orquestra.quantum.openfermion import QubitOperator

from orquestra.opt.problems import solve_problem_by_exhaustive_search

HAMILTONIAN_SOLUTION_COST_LIST = [
    (
        QubitOperator("Z0 Z1") + QubitOperator("[]", -1),
        [(0, 1), (1, 0)],
        -2,
    ),
    (
        QubitOperator("Z0 Z1", 5)
        + QubitOperator("Z0 Z3", 5)
        + QubitOperator("Z1 Z2", 0.5)
        + QubitOperator("Z2 Z3", 0.5)
        + QubitOperator("[]", -11),
        [(0, 1, 0, 1), (1, 0, 1, 0)],
        -22,
    ),
]


class TestSolveProblemByExhaustiveSearch:
    @pytest.mark.parametrize(
        "hamiltonian,target_solutions,target_value", [*HAMILTONIAN_SOLUTION_COST_LIST]
    )
    def test_solve_problem_by_exhaustive_search(
        self, hamiltonian, target_solutions, target_value
    ):
        value, solutions = solve_problem_by_exhaustive_search(hamiltonian)
        assert set(solutions) == set(target_solutions)
        assert value == target_value



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import copy

import pytest

from orquestra.opt.problems import MaxCut

from ._helpers import make_graph

MONOTONIC_GRAPH_OPERATOR_TERM_PAIRS = [
    (
        make_graph(node_ids=range(2), edges=[(0, 1)]),
        {
            (): -0.5,
            ((0, "Z"), (1, "Z")): 0.5,
        },
    ),
    (
        make_graph(node_ids=range(3), edges=[(0, 1), (0, 2)]),
        {
            (): -1,
            ((0, "Z"), (1, "Z")): 0.5,
            ((0, "Z"), (2, "Z")): 0.5,
        },
    ),
    (
        make_graph(
            node_ids=range(4), edges=[(0, 1, 1), (0, 2, 2), (0, 3, 3)], use_weights=True
        ),
        {
            (): -3,
            ((0, "Z"), (1, "Z")): 0.5,
            ((0, "Z"), (2, "Z")): 1,
            ((0, "Z"), (3, "Z")): 1.5,
        },
    ),
    (
        make_graph(node_ids=range(5), edges=[(0, 1), (1, 2), (3, 4)]),
        {
            (): -1.5,
            ((0, "Z"), (1, "Z")): 0.5,
            ((1, "Z"), (2, "Z")): 0.5,
            ((3, "Z"), (4, "Z")): 0.5,
        },
    ),
]

GRAPH_OPERATOR_TERM_SCALING_OFFSET_LIST = [
    (
        make_graph(
            node_ids=range(4),
            edges=[(0, 1, 1), (0, 2, 2), (0, 3, 3)],
            use_weights=True,
        ),
        {
            (): 1,
            ((0, "Z"), (1, "Z")): 1,
            ((0, "Z"), (2, "Z")): 2,
            ((0, "Z"), (3, "Z")): 3,
        },
        2.0,
        7.0,
    ),
]

NONMONOTONIC_GRAPH_OPERATOR_TERM_PAIRS = [
    (
        make_graph(node_ids=[4, 2], edges=[(2, 4)]),
        {
            (): -0.5,
            ((0, "Z"), (1, "Z")): 0.5,
        },
    ),
    (
        make_graph(node_ids="CBA", edges=[("C", "B"), ("C", "A")]),
        {
            (): -1,
            ((0, "Z"), (1, "Z")): 0.5,  # the C-B edge
            ((0, "Z"), (2, "Z")): 0.5,  # the C-A edge
        },
    ),
]

GRAPH_SOLUTION_COST_LIST = [
    (make_graph(node_ids=range(2), edges=[(0, 1)]), [0, 0], 0),
    (make_graph(node_ids=range(2), edges=[(0, 1)]), [0, 1], -1),
    (
        make_graph(
            node_ids=range(4), edges=[(0, 1, 1), (0, 2, 2), (0, 3, 3)], use_weights=True
        ),
        [1, 0, 0, 0],
        -6,
    ),
    (make_graph(node_ids=range(4), edges=[(0, 1), (0, 2), (0, 3)]), [0, 0, 1, 1], -2),
    (make_graph(node_ids=range(4), edges=[(0, 1), (0, 2), (0, 3)]), [0, 1, 1, 1], -3),
    (
        make_graph(node_ids=range(5), edges=[(0, 1), (1, 2), (3, 4)]),
        [1, 1, 1, 1, 1],
        0,
    ),
]

GRAPH_BEST_SOLUTIONS_COST_LIST = [
    (make_graph(node_ids=range(2), edges=[(0, 1)]), [(0, 1), (1, 0)], -1),
    (
        make_graph(node_ids=range(3), edges=[(0, 1), (0, 2)]),
        [(0, 1, 1), (1, 0, 0)],
        -2,
    ),
    (
        make_graph(node_ids=range(4), edges=[(0, 1), (0, 2), (0, 3)]),
        [
            (0, 1, 1, 1),
            (1, 0, 0, 0),
        ],
        -3,
    ),
    (
        make_graph(node_ids=range(5), edges=[(0, 1), (1, 2), (3, 4)]),
        [(0, 1, 0, 0, 1), (0, 1, 0, 1, 0), (1, 0, 1, 0, 1), (1, 0, 1, 1, 0)],
        -3,
    ),
]


class TestGetMaxcutHamiltonian:
    @pytest.mark.parametrize(
        "graph,terms",
        [
            *MONOTONIC_GRAPH_OPERATOR_TERM_PAIRS,
            *NONMONOTONIC_GRAPH_OPERATOR_TERM_PAIRS,
        ],
    )
    def test_returns_expected_terms(self, graph, terms):
        qubit_operator = MaxCut().get_hamiltonian(graph)
        assert qubit_operator.terms == terms

    @pytest.mark.parametrize(
        "graph,terms,scale_factor,offset",
        [*GRAPH_OPERATOR_TERM_SCALING_OFFSET_LIST],
    )
    def test_scaling_and_offset_works(self, graph, terms, scale_factor, offset):
        qubit_operator = MaxCut().get_hamiltonian(graph, scale_factor, offset)
        assert qubit_operator.terms == terms


class TestEvaluateMaxcutSolution:
    @pytest.mark.parametrize("graph,solution,target_value", [*GRAPH_SOLUTION_COST_LIST])
    def test_evaluate_maxcut_solution(self, graph, solution, target_value):
        value = MaxCut().evaluate_solution(solution, graph)
        assert value == target_value

    @pytest.mark.parametrize("graph,solution,target_value", [*GRAPH_SOLUTION_COST_LIST])
    def test_evaluate_maxcut_solution_with_invalid_input(
        self, graph, solution, target_value
    ):
        too_long_solution = solution + [1]
        too_short_solution = solution[:-1]
        invalid_value_solution = copy.copy(solution)
        invalid_value_solution[0] = -1
        invalid_solutions = [
            too_long_solution,
            too_short_solution,
            invalid_value_solution,
        ]
        for invalid_solution in invalid_solutions:
            with pytest.raises(ValueError):
                _ = MaxCut().evaluate_solution(invalid_solution, graph)


class TestSolveMaxcutByExhaustiveSearch:
    @pytest.mark.parametrize(
        "graph,target_solutions,target_value", [*GRAPH_BEST_SOLUTIONS_COST_LIST]
    )
    def test_solve_maxcut_by_exhaustive_search(
        self, graph, target_solutions, target_value
    ):
        value, solutions = MaxCut().solve_by_exhaustive_search(graph)
        assert set(solutions) == set(target_solutions)
        assert value == target_value



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from .basin_hopping import BasinHoppingOptimizer
from .cma_es_optimizer import CMAESOptimizer
from .qiskit_optimizer import QiskitOptimizer
from .scipy_optimizer import ScipyOptimizer
from .search_points_optimizer import SearchPointsOptimizer
from .simple_gradient_descent import SimpleGradientDescent



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""Interfaces related to cost functions."""
from typing import Union

import numpy as np
from orquestra.quantum.utils import ValueEstimate
from typing_extensions import Protocol

from .functions import (
    CallableStoringArtifacts,
    CallableWithGradient,
    CallableWithGradientStoringArtifacts,
)


class _CostFunction(Protocol):
    """Cost function transforming vectors from R^n to numbers or their estimates."""

    def __call__(self, parameters: np.ndarray) -> Union[float, ValueEstimate]:
        """Compute  value of the cost function for given parameters."""
        ...


CostFunction = Union[
    _CostFunction,
    CallableWithGradient,
    CallableStoringArtifacts,
    CallableWithGradientStoringArtifacts,
]


class ParameterPreprocessor(Protocol):
    """Parameter preprocessor.

    Implementer of this protocol should create new array instead of
    modifying passed parameters in place, which can have unpredictable
    side effects.
    """

    def __call__(self, parameters: np.ndarray) -> np.ndarray:
        """Preprocess parameters."""
        ...



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import numpy as np

from orquestra.opt.mock_objects import MockOptimizer


class CustomMockOptimizer(MockOptimizer):
    def _preprocess_cost_function(self, cost_function):
        def _new_cost_function(parameters):
            return cost_function(parameters) + 1

        return _new_cost_function


def simple_cost_function(parameters):
    return 1


def test_cost_function_is_preprocessed_before_minimization():
    optimizer = CustomMockOptimizer()
    result = optimizer.minimize(simple_cost_function, np.zeros(4))

    assert result.opt_value == 2



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import networkx as nx


def make_graph(node_ids, edges, use_weights=False):
    graph = nx.Graph()
    graph.add_nodes_from(node_ids)
    if use_weights:
        graph.add_weighted_edges_from(edges)
    else:
        graph.add_edges_from(edges)
    return graph


def graph_node_index(graph, node_id):
    return next(node_i for node_i, node in enumerate(graph.nodes) if node == node_id)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from .cost_function import CostFunction
from .functions import (
    CallableStoringArtifacts,
    CallableWithGradient,
    CallableWithGradientStoringArtifacts,
    FunctionWithGradient,
    FunctionWithGradientStoringArtifacts,
    StoreArtifact,
    function_with_gradient,
    has_store_artifact_param,
)
from .optimizer import (
    NestedOptimizer,
    Optimizer,
    construct_history_info,
    optimization_result,
)
from .save_conditions import SaveCondition, always, every_nth



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from typing import Optional, Tuple

import cvxpy as cp
import numpy as np
from scipy.optimize import LinearConstraint
from typing_extensions import Protocol, runtime_checkable

from .api import Optimizer


def solve_qp_problem_for_psd_matrix(
    matrix: np.ndarray, symmetrize: bool = True
) -> Tuple[np.ndarray, float]:
    """
    Solves a quadratic programming (QP) optimization problem. The matrix should
    be positive semi-definite. This implementation assumes that the domain of
    the solution are variables between 0 and 1. If matrix is not positive
    semi-definite, `solve_qp_problem_with_optimizer` method should be used.

    Notes:
        We are aware of the fact that in the specified domain the solution is
        always a vector of zeros, but decided to leave the implementation as it is
        in case the domain changes in future.

    Args:
        matrix: a matrix representing the problem.
        symmetrize: a flag indicating whether the matrix should be symmetrized.

    Returns:
        np.ndarray: vector representing solution to the problem.
        float: optimal value of the solution.
    """
    if symmetrize:
        matrix = (matrix + matrix.T) / 2

    if not is_matrix_positive_semidefinite(matrix):
        raise ValueError("Input matrix should be positive semi-definite.")

    size = matrix.shape[0]
    P = matrix
    G = np.vstack([np.eye(size), -np.eye(size)])
    h = np.hstack([np.ones(size), np.zeros(size)])

    x = cp.Variable(size)
    problem = cp.Problem(cp.Minimize((1 / 2) * cp.quad_form(x, P)), [G @ x <= h])

    problem.solve()
    return x.value, problem.value


def solve_qp_problem_with_optimizer(
    matrix: np.ndarray,
    optimizer: Optimizer,
    number_of_trials: int = 1,
    symmetrize: bool = True,
) -> Tuple[np.ndarray, float]:
    """
    Solves a quadratic programming (QP) optimization problem.
    This implementation assumes that the domain of the solution are variables
    between 0 and 1.

    Args:
        matrix: a matrix representing the problem.
        optimizer: an optimizer to be used to solve the problem. Optimizer should
            support constraints.
        number_of_trials: specifies the number of times problem will be solved.
            Only the best solution will be returned.
        symmetrize: a flag indicating whether the matrix should be symmetrized.

    Returns:
        np.ndarray: vector representing solution to the problem
        float: optimal value of the solution
    """
    if symmetrize:
        matrix = (matrix + matrix.T) / 2

    # Use an optimizer to solve non-convex QP relaxations
    if not isinstance(optimizer, _WithConstraints):
        raise ValueError("Optimizer needs to support constraints.")
    size = matrix.shape[0]
    A = np.eye(size)
    lower_bound = np.zeros(size)
    upper_bound = np.ones(size)
    linear_constraint = LinearConstraint(A, lower_bound, upper_bound)

    optimizer.constraints = linear_constraint

    def cost_function(x):
        return x.T @ matrix @ x

    final_value = None
    final_params = None

    for _ in range(number_of_trials):
        initial_params = np.random.uniform(0.0, 1.0, size=size)
        optimization_results = optimizer.minimize(cost_function, initial_params)
        if final_value is None or optimization_results.opt_value < final_value:
            final_value = optimization_results.opt_value
            final_params = optimization_results.opt_params
    assert final_value is not None
    assert final_params is not None
    # We round the values to avoid having values like 1.0000000002 or -1e-14
    # in the output
    return np.around(final_params, decimals=8), final_value


def is_matrix_positive_semidefinite(matrix: np.ndarray) -> bool:
    """
    Checks whether matrix is positive semi-definite.

    Args:
        matrix: a matrix that should be checked.

    Returns:
        bool: True if matrix is positive semi-definite, False otherwise.

    """
    eigenvalues, _ = np.linalg.eig(matrix)
    return np.min(eigenvalues) >= 0


# Temporary solution until constraints added to optimizer.py in orquestra-opt
@runtime_checkable
class _WithConstraints(Protocol):
    """Minimum interface for Optimizers with constraints"""

    constraints: Optional[LinearConstraint]



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import numpy as np
import pytest

from orquestra.opt.gradients import finite_differences_gradient


def sum_x_squared(parameters: np.ndarray) -> float:
    return (parameters**2).sum()


@pytest.mark.parametrize(
    "parameters", [np.array([1, 2, 3]), np.array([0.5, 0.25, 1, 0.0])]
)
def test_finite_differences_gradient_returns_vectors_with_correct_length(parameters):
    gradient = finite_differences_gradient(sum_x_squared)

    assert len(parameters) == len(gradient(parameters))


@pytest.mark.parametrize(
    "epsilon,parameters",
    [
        (0.1, np.array([0, 0, 0])),
        (0.001, np.array([0, 1, 0])),
        (0.001, np.array([-0.5, 0.25, 1])),
    ],
)
def test_finite_differences_gradient_uses_supplied_epsilon_to_compute_gradient_estimate(
    epsilon, parameters
):
    gradient = finite_differences_gradient(sum_x_squared, epsilon)
    eps_vectors = np.eye(len(parameters)) * epsilon

    expected_gradient_value = np.array(
        [
            sum_x_squared(parameters + vector) - sum_x_squared(parameters - vector)
            for vector in eps_vectors
        ]
    ) / (2 * epsilon)

    assert np.array_equal(expected_gradient_value, gradient(parameters))



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from typing import Callable, Union

import numpy as np
import scipy.optimize

from ..api import (
    CallableWithGradient,
    Optimizer,
    construct_history_info,
    optimization_result,
)
from ..history.recorder import RecorderFactory
from ..history.recorder import recorder as _recorder


class BasinHoppingOptimizer(Optimizer):
    def __init__(
        self,
        niter: int = 100,
        T: float = 1.0,
        stepsize: float = 0.5,
        minimizer_kwargs: Union[dict, None] = None,
        take_step: Union[Callable, None] = None,
        accept_test: Union[Callable, None] = None,
        interval: int = 50,
        disp: bool = False,
        niter_success: Union[int, None] = None,
        recorder: RecorderFactory = _recorder,
    ):
        """The BasinHoppingOptimizer utilizes the scipy.optimize.basinhopping method
        (https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.basinhopping.html).
        It is intended to be used in conjunction with methods of
        scipy.optimize.minimize for local optimization.

        Args:
            niter: See scipy.optimize.basinhopping
            T: See scipy.optimize.basinhopping
            stepsize: See scipy.optimize.basinhopping
            minimizer_kwargs: See scipy.optimize.basinhopping
            take_step: See scipy.optimize.basinhopping
            accept_test: See scipy.optimize.basinhopping
            interval: See scipy.optimize.basinhopping
            disp: See scipy.optimize.basinhopping
            niter_success: See scipy.optimize.basinhopping
        """  # noqa
        super().__init__(recorder=recorder)
        self.niter = niter
        self.T = T
        self.stepsize = stepsize
        self.minimizer_kwargs = minimizer_kwargs
        self.take_step = take_step
        self.accept_test = accept_test
        self.interval = interval
        self.disp = disp
        self.niter_success = niter_success

    def _minimize(
        self,
        cost_function: Union[CallableWithGradient, Callable],
        initial_params: np.ndarray = None,
        keep_history: bool = False,
    ):
        """
        Minimizes given cost function using functions from scipy.optimize.basinhopping.

        Args:
            cost_function(): python method which takes numpy.ndarray as input
            initial_params(np.ndarray): initial parameters to be used for optimization
            keep_history: flag indicating whether history of cost function
                evaluations should be recorded.
        """
        jacobian = None
        if isinstance(cost_function, CallableWithGradient) and callable(
            getattr(cost_function, "gradient")
        ):
            jacobian = cost_function.gradient
        if (
            self.minimizer_kwargs is not None
            and self.minimizer_kwargs.get("options", None) is not None
        ):
            self.minimizer_kwargs["options"]["jacobian"] = jacobian

        result = scipy.optimize.basinhopping(
            cost_function,
            initial_params,
            niter=self.niter,
            T=self.T,
            stepsize=self.stepsize,
            minimizer_kwargs=self.minimizer_kwargs,
            take_step=self.take_step,
            accept_test=self.accept_test,
            interval=self.interval,
            disp=self.disp,
            niter_success=self.niter_success,
        )

        opt_value = result.fun
        opt_params = result.x

        nit = result.get("nit", None)
        nfev = result.get("nfev", None)

        return optimization_result(
            opt_value=opt_value,
            opt_params=opt_params,
            nit=nit,
            nfev=nfev,
            **construct_history_info(cost_function, keep_history)  # type: ignore
        )



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from typing import Callable, List, Optional, Union

import numpy as np
from scipy.optimize import OptimizeResult

from ..api import (
    CallableWithGradient,
    Optimizer,
    construct_history_info,
    optimization_result,
)
from ..history.recorder import RecorderFactory
from ..history.recorder import recorder as _recorder


class SearchPointsOptimizer(Optimizer):
    def __init__(
        self,
        parameter_values_list: List[np.ndarray],
        recorder: RecorderFactory = _recorder,
    ):
        """
        Args:
            parameter_values_list: list of parameter values to evaluate
            recorder: recorder object which defines how to store
                the optimization history.
        """
        super().__init__(recorder=recorder)
        assert len(parameter_values_list) > 0
        self.parameter_values_list = parameter_values_list

    def _minimize(
        self,
        cost_function: Union[CallableWithGradient, Callable],
        initial_params: Optional[np.ndarray] = None,
        keep_history: bool = False,
    ) -> OptimizeResult:
        """
        Finds the parameters which minimize given cost function, by trying
        all the parameters from the provided list of points.

        Args:
            cost_function: object representing cost function we want to minimize
            inital_params: initial parameters for the cost function
            keep_history: flag indicating whether history of cost function
                evaluations should be recorded.

        """
        if initial_params is not None and len(initial_params) != 0:
            Warning(
                "DiscreteParameterValuesSearch search doesn't use initial parameters, "
                "they will be ignored."
            )

        min_value = None
        optimal_params = None

        for parameter_values in self.parameter_values_list:
            value = cost_function(parameter_values)
            if min_value is None or value < min_value:
                min_value = value
                optimal_params = parameter_values

        return optimization_result(
            opt_value=min_value,
            opt_params=optimal_params,
            nfev=len(self.parameter_values_list),
            nit=None,
            **construct_history_info(cost_function, keep_history)  # type: ignore
        )



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from typing import Callable, Dict, List

import networkx as nx
import numpy as np
from orquestra.quantum.openfermion import IsingOperator, QubitOperator


def get_random_ising_hamiltonian(
    number_of_qubits: int, number_of_terms: int, max_number_of_qubits_per_term: int
) -> IsingOperator:
    """Generates a random Hamiltonian for a given number of qubits and terms with
    weights between -1 and 1.

    NOTE: Due to randomness, we cannot ensure that the returned hamiltonian has an
        operation on every qubit.

    Args:
        number_of_qubits: The number of qubits in the Hamiltonian. Should be >= 2.
        max_number_qubits_per_term: The maximum number of qubits for each term in the
            hamiltonian. Should be <= number_of_qubits.
    """
    # Initiate hamiltonian with a random coefficient
    hamiltonian = IsingOperator("", np.random.rand() * 2 - 1)

    # Add terms with random qubits
    for _ in range(number_of_terms):
        num_qubits_in_term = np.random.randint(1, max_number_of_qubits_per_term + 1)
        qubits = np.random.choice(
            range(number_of_qubits), num_qubits_in_term, replace=False
        )
        hamiltonian += IsingOperator(
            " ".join([f"Z{q}" for q in qubits]), np.random.rand() * 2 - 1
        )

    return hamiltonian



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import numpy as np

from orquestra.opt.api import FunctionWithGradient
from orquestra.opt.api.optimizer import construct_history_info
from orquestra.opt.gradients import finite_differences_gradient
from orquestra.opt.history import recorder


def sum_x_squared(x):
    return sum(x**2.0)


class TestConstructHistoryInfo:
    def test_history_is_empty_if_keep_value_history_is_false(self):
        cost_function = recorder(sum_x_squared)

        cost_function(np.array([1, 2, 3]))

        history_info = construct_history_info(cost_function, False)

        assert not history_info["history"]
        assert "gradient_history" not in history_info

    def test_history_info_contains_only_history_for_function_without_gradient(self):
        cost_function = recorder(sum_x_squared)

        cost_function(np.array([1, 2, 3]))

        history_info = construct_history_info(cost_function, True)

        assert len(history_info["history"]) == 1
        assert "gradient_history" not in history_info

        history_entry = history_info["history"][0]
        assert history_entry.call_number == 0
        np.testing.assert_array_equal(history_entry.params, [1, 2, 3])
        assert history_entry.value == cost_function(np.array([1, 2, 3]))

    def test_history_info_contains_gradient_history_for_function_with_gradient(self):
        cost_function = recorder(
            FunctionWithGradient(
                sum_x_squared, finite_differences_gradient(sum_x_squared)
            )
        )

        cost_function(np.array([1, 2, 3]))
        cost_function.gradient(np.array([0, -1, 1]))

        history_info = construct_history_info(cost_function, True)

        assert len(history_info["history"]) == 1
        assert len(history_info["gradient_history"]) == 1

        history_entry = history_info["history"][0]
        assert history_entry.call_number == 0
        np.testing.assert_array_equal(history_entry.params, [1, 2, 3])
        assert history_entry.value == cost_function(np.array([1, 2, 3]))

        history_entry = history_info["gradient_history"][0]
        assert history_entry.call_number == 0
        np.testing.assert_array_equal(history_entry.params, [0, -1, 1])
        np.testing.assert_array_equal(
            history_entry.value, cost_function.gradient(np.array([0, -1, 1]))
        )



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from typing import Callable, Dict, Optional, Union

import numpy as np
from qiskit.algorithms.optimizers import ADAM, NFT, SPSA
from scipy.optimize import OptimizeResult

from ..api import (
    CallableWithGradient,
    Optimizer,
    construct_history_info,
    optimization_result,
)
from ..history.recorder import RecorderFactory
from ..history.recorder import recorder as _recorder


class QiskitOptimizer(Optimizer):
    def __init__(
        self,
        method: str,
        optimizer_kwargs: Optional[Dict] = None,
        recorder: RecorderFactory = _recorder,
    ):
        """
        Args:
            method: specifies optimizer to be used.
                Currently supports "ADAM", "AMSGRAD" and "SPSA".
            optimizer_kwargs: dictionary with additional optimizer_kwargs
                for the optimizer.
            recorder: recorder object which defines how to store
                the optimization history.

        """
        super().__init__(recorder=recorder)
        self.method = method
        if optimizer_kwargs is None:
            self.optimizer_kwargs = {}
        else:
            self.optimizer_kwargs = optimizer_kwargs

        if self.method == "SPSA":
            self.optimizer = SPSA(**self.optimizer_kwargs)
        elif self.method == "ADAM" or self.method == "AMSGRAD":
            if self.method == "AMSGRAD":
                self.optimizer_kwargs["amsgrad"] = True
            self.optimizer = ADAM(**self.optimizer_kwargs)
        elif self.method == "NFT":
            self.optimizer = NFT(**self.optimizer_kwargs)

    def _minimize(
        self,
        cost_function: Union[CallableWithGradient, Callable],
        initial_params: np.ndarray,
        keep_history: bool = False,
    ) -> OptimizeResult:
        """
        Minimizes given cost function using optimizers from Qiskit Aqua.

        Args:
            cost_function: python method which takes numpy.ndarray as input
            initial_params: initial parameters to be used for optimization

        Returns:
            optimization_results: results of the optimization.
        """

        number_of_variables = len(initial_params)

        gradient_function = None
        if isinstance(cost_function, CallableWithGradient):
            gradient_function = cost_function.gradient

        solution, value, nfev = self.optimizer.optimize(
            num_vars=number_of_variables,
            objective_function=cost_function,
            initial_point=initial_params,
            gradient_function=gradient_function,
        )

        if self.method == "ADAM" or self.method == "AMSGRAD":
            nit = self.optimizer._t
        elif self.method == "NFT":
            nit = self.optimizer._options["maxiter"]
        else:
            nit = self.optimizer.maxiter

        return optimization_result(
            opt_value=value,
            opt_params=solution,
            nit=nit,
            nfev=nfev,
            **construct_history_info(cost_function, keep_history)
        )



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import dimod


class BQMSolver(dimod.Sampler):
    """
    This method is right now just a trivial wrapper around dimod.Sampler .
    The main reason for its existence is to make it easier to keep backward
    compatibility.
    If we ever implement some additional functionalities for this class it
    won't require changing the base class of the already implemented samplers.
    """

    def solve(self, bqm: dimod.BQM) -> dimod.SampleSet:
        """
        Wrapper around the `sample` method of dimod.Sampler, for making it more
        intuitive to use, as some people find the term "sampling" confusing in
        the context of combinatorial optimization problems.

        Note:
            This method should not be overwritten by inheriting classes as it's
            intended only as a utility or "thin wrapper". Any logic should be included
            in the `sample`, `sample_qubo` or `sample_ising` methods to keep
            compatibiltity with the `dimod.Sampler` class.

        """
        return self.sample(bqm)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from copy import copy, deepcopy

import pytest

from orquestra.opt.api.example_functions import (
    Function2,
    Function5,
    function_1,
    function_3,
    function_4,
    function_6,
)
from orquestra.opt.history import recorder


@pytest.mark.parametrize("func", [Function2(5), function_3, function_4, Function5(0.5)])
def test_recorder_can_be_copied_shallowly(func):
    recorded = recorder(func)

    recorded_copy = copy(recorded)

    assert recorded_copy.target is recorded.target
    assert recorded_copy.predicate is recorded.predicate


@pytest.mark.parametrize("func", [Function2(5), function_3, function_4, Function5(0.5)])
def test_recorder_can_be_copied_deeply(func):
    recorded = recorder(func)

    recorded_copy = deepcopy(recorded)

    assert recorded_copy.target == recorded.target
    assert recorded_copy.predicate == recorded.predicate



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import pytest

from orquestra.opt.api.optimizer_test import OptimizerTests
from orquestra.opt.optimizers.basin_hopping import BasinHoppingOptimizer


@pytest.fixture(
    params=[
        {
            "niter": 3,
        },
        {
            "niter": 3,
            "minimizer_kwargs": {
                "method": "BFGS",
            },
        },
        {
            "niter": 3,
            "minimizer_kwargs": {
                "method": "L-BFGS-B",
                "options": {
                    "ftol": 1e-7,
                },
            },
        },
        {
            "niter": 3,
            "minimizer_kwargs": {
                "method": "Nelder-Mead",
                "options": {
                    "fatol": 1e-7,
                },
            },
        },
        {
            "niter": 3,
            "minimizer_kwargs": {
                "method": "SLSQP",
                "options": {
                    "ftol": 1e-7,
                },
            },
        },
        {
            "niter": 3,
            "minimizer_kwargs": {
                "method": "COBYLA",
                "options": {"maxiter": 1000000, "tol": 1e-7},
            },
        },
    ]
)
def optimizer(request):
    return BasinHoppingOptimizer(**request.param)


@pytest.fixture(params=[True, False])
def keep_history(request):
    return request.param


class TestBasinHoppingOptimizer(OptimizerTests):
    pass



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""Test cases for recording functions with gradients."""
import numpy as np
import pytest

from orquestra.opt.api import CallableWithGradient
from orquestra.opt.api.example_functions import Function2, Function5, function_1
from orquestra.opt.history import recorder


@pytest.mark.parametrize(
    "function,params",
    [
        (function_1, np.array([3, 4])),
        (Function2(5), np.array([-1, 0, 1])),
        (Function5(10), np.array([1, 2, 3])),
    ],
)
def test_recorder_propagates_calls_to_wrapped_functions_and_its_gradient(
    function: CallableWithGradient, params: np.ndarray
):
    target = recorder(function)
    assert target(params) == function(params)
    np.testing.assert_array_almost_equal(
        target.gradient(params), function.gradient(params)
    )


@pytest.mark.parametrize(
    "function,params",
    [
        (function_1, np.array([3, 4])),
        (Function2(5), np.array([-1, 0, 1])),
        (Function5(10), np.array([1, 2, 3])),
    ],
)
def test_calls_to_gradient_are_also_recorded(
    function: CallableWithGradient, params: np.ndarray
):
    target = recorder(function)
    np.testing.assert_array_equal(function.gradient(params), target.gradient(params))

    assert len(target.gradient.history) == 1
    entry = target.gradient.history[0]

    assert entry.call_number == 0
    np.testing.assert_array_equal(entry.params, params)
    np.testing.assert_array_equal(entry.value, target.gradient(params))
